"""Module for the Neural model classes."""

__all__ = [
    "FeedForward",
    "ResidualFeedForward",
    "MultiFeedForward",
    "DeepONet",
    "MIONet",
    "FNO",
    "FourierIntegralKernel",
    "KernelNeuralOperator",
    "AveragingNeuralOperator",
    "LowRankNeuralOperator",
    "Spline",
    "GraphNeuralOperator",
    "PirateNet",
    "EquivariantGraphNeuralOperator",
    "SINDy",
]

from .feed_forward import FeedForward, ResidualFeedForward
from .multi_feed_forward import MultiFeedForward
from .deeponet import DeepONet, MIONet
from .fourier_neural_operator import FNO, FourierIntegralKernel
from .kernel_neural_operator import KernelNeuralOperator
from .average_neural_operator import AveragingNeuralOperator
from .low_rank_neural_operator import LowRankNeuralOperator
from .spline import Spline
from .spline_surface import SplineSurface
from .graph_neural_operator import GraphNeuralOperator
from .pirate_network import PirateNet
from .equivariant_graph_neural_operator import EquivariantGraphNeuralOperator
from .sindy import SINDy



"""Module for the Weighting Interface."""

from abc import ABCMeta, abstractmethod
from typing import final
from ..utils import check_positive_integer, is_function

_AGGREGATE_METHODS = {"sum": sum, "mean": lambda x: sum(x) / len(x)}


class WeightingInterface(metaclass=ABCMeta):
    """
    Abstract base class for all loss weighting schemas. All weighting schemas
    should inherit from this class.
    """

    def __init__(self, update_every_n_epochs=1, aggregator="sum"):
        """
        Initialization of the :class:`WeightingInterface` class.

        :param int update_every_n_epochs: The number of training epochs between
            weight updates. If set to 1, the weights are updated at every epoch.
            This parameter is ignored by static weighting schemes. Default is 1.
        :param aggregator: The aggregation method. Either:
            - 'sum' → torch.sum
            - 'mean' → torch.mean
            - callable → custom aggregation function
        :type aggregator: str | Callable
        """
        # Check consistency
        check_positive_integer(value=update_every_n_epochs, strict=True)

        # Aggregation
        if isinstance(aggregator, str):
            if aggregator not in _AGGREGATE_METHODS:
                raise ValueError(
                    f"Invalid aggregator '{aggregator}'. Must be one of "
                    f"{list(_AGGREGATE_METHODS.keys())}."
                )
            aggregator = _AGGREGATE_METHODS[aggregator]

        elif not is_function(aggregator):
            raise TypeError(
                f"Aggregator must be either a string or a callable, "
                f"got {type(aggregator).__name__}."
            )

        # Initialization
        self._solver = None
        self.update_every_n_epochs = update_every_n_epochs
        self.aggregator_fn = aggregator
        self._saved_weights = {}

    @abstractmethod
    def weights_update(self, losses):
        """
        Update the weighting scheme based on the given losses.

        This method must be implemented by subclasses. Its role is to update the
        values of the weights. The updated weights will then be used by
        :meth:`aggregate` to compute the final aggregated loss.

        :param dict losses: The dictionary of losses.
        :return: The updated weights.
        :rtype: dict
        """

    @final
    def aggregate(self, losses):
        """
        Update the weights (if needed) and aggregate the given losses.

        This method first checks whether the loss weights need to be updated
        based on the current epoch and the ``update_every_n_epochs`` setting.
        If an update is required, it calls :meth:`weights_update` to refresh the
        weights. Afterwards, it aggregates the (weighted) losses into a single
        scalar tensor using the configured aggregator function. This method must
        not be overridden.

        :param dict losses: The dictionary of losses.
        :return: The aggregated loss tensor.
        :rtype: torch.Tensor
        """
        # Update weights
        if self.solver.trainer.current_epoch % self.update_every_n_epochs == 0:
            self._saved_weights = self.weights_update(losses)

        # Aggregate. Using direct indexing instead of .get() ensures that a
        # KeyError is raised if the expected condition is missing from the dict.
        return self.aggregator_fn(
            self._saved_weights[condition] * loss
            for condition, loss in losses.items()
        )

    def last_saved_weights(self):
        """
        Get the last saved weights.

        :return: The last saved weights.
        :rtype: dict
        """
        return self._saved_weights

    @property
    def solver(self):
        """
        The solver employing this weighting schema.

        :return: The solver.
        :rtype: :class:`~pina.solver.SolverInterface`
        """
        return self._solver



"""Module for the DomainEquationCondition class."""

from .condition_interface import ConditionInterface
from ..utils import check_consistency
from ..domain import DomainInterface
from ..equation.equation_interface import EquationInterface


class DomainEquationCondition(ConditionInterface):
    """
    The class :class:`DomainEquationCondition` defines a condition based on a
    ``domain`` and an ``equation``. This condition is typically used in
    physics-informed problems, where the model is trained to satisfy a given
    ``equation`` over a specified ``domain``. The ``domain`` is used to sample
    points where the ``equation`` residual is evaluated and minimized during
    training.

    :Example:

    >>> from pina.domain import CartesianDomain
    >>> from pina.equation import Equation
    >>> from pina import Condition

    >>> # Equation to be satisfied over the domain: # x^2 + y^2 - 1 = 0
    >>> def dummy_equation(pts):
    ...     return pts["x"]**2 + pts["y"]**2 - 1

    >>> domain = CartesianDomain({"x": [0, 1], "y": [0, 1]})
    >>> condition = Condition(domain=domain, equation=Equation(dummy_equation))
    """

    # Available slots
    __slots__ = ["domain", "equation"]

    def __init__(self, domain, equation):
        """
        Initialization of the :class:`DomainEquationCondition` class.

        :param DomainInterface domain: The domain over which the equation is
            defined.
        :param EquationInterface equation: The equation to be satisfied over the
            specified domain.
        """
        super().__init__()
        self.domain = domain
        self.equation = equation

    def __setattr__(self, key, value):
        """
        Set the attribute value with type checking.

        :param str key: The attribute name.
        :param any value: The value to set for the attribute.
        """
        if key == "domain":
            check_consistency(value, (DomainInterface, str))
            DomainEquationCondition.__dict__[key].__set__(self, value)

        elif key == "equation":
            check_consistency(value, (EquationInterface))
            DomainEquationCondition.__dict__[key].__set__(self, value)

        elif key in ("_problem"):
            super().__setattr__(key, value)



"""Module for the SINDy model class."""

from typing import Callable
import torch
from ..utils import check_consistency, check_positive_integer


class SINDy(torch.nn.Module):
    r"""
    SINDy model class.

    The Sparse Identification of Nonlinear Dynamics (SINDy) model identifies the
    governing equations of a dynamical system from data by learning a sparse
    linear combination of non-linear candidate functions.

    The output of the model is expressed as product of a library matrix and a
    coefficient matrix:

    .. math::

        \dot{X} = \Theta(X) \Xi

    where:
      - :math:`X \in \mathbb{R}^{B \times D}` is the input snapshots of the
        system state. Here, :math:`B` is the batch size and :math:`D` is the
        number of state variables.
      - :math:`\Theta(X) \in \mathbb{R}^{B \times L}` is the library matrix
        obtained by evaluating a set of candidate functions on the input data.
        Here, :math:`L` is the number of candidate functions in the library.
      - :math:`\Xi \in \mathbb{R}^{L \times D}` is the learned coefficient
        matrix that defines the sparse model.

    .. seealso::

        **Original reference**:
        Brunton, S.L., Proctor, J.L., and Kutz, J.N. (2016).
        *Discovering governing equations from data: Sparse identification of
        non-linear dynamical systems.*
        Proceedings of the National Academy of Sciences, 113(15), 3932-3937.
        DOI: `10.1073/pnas.1517384113
        <https://doi.org/10.1073/pnas.1517384113>`_
    """

    def __init__(self, library, output_dimension):
        """
        Initialization of the :class:`SINDy` class.

        :param list[Callable] library: The collection of candidate functions
            used to construct the library matrix. Each function must accept an
            input tensor of shape ``[..., D]`` and return a tensor of shape
            ``[..., 1]``.
        :param int output_dimension: The number of output variables, typically
            the number of state derivatives. It determines the number of columns
            in the coefficient matrix.
        :raises ValueError: If ``library`` is not a list of callables.
        :raises AssertionError: If ``output_dimension`` is not a positive
            integer.
        """
        super().__init__()

        # Check consistency
        check_positive_integer(output_dimension, strict=True)
        check_consistency(library, Callable)
        if not isinstance(library, list):
            raise ValueError("`library` must be a list of callables.")

        # Initialization
        self._library = library
        self._coefficients = torch.nn.Parameter(
            torch.zeros(len(library), output_dimension)
        )

    def forward(self, x):
        """
        Forward pass of the :class:`SINDy` model.

        :param torch.Tensor x: The input batch of state variables.
        :return: The predicted time derivatives of the state variables.
        :rtype: torch.Tensor
        """
        theta = torch.stack([f(x) for f in self.library], dim=-2)
        return torch.einsum("...li , lo -> ...o", theta, self.coefficients)

    @property
    def library(self):
        """
        The library of candidate functions.

        :return: The library.
        :rtype: list[Callable]
        """
        return self._library

    @property
    def coefficients(self):
        """
        The coefficients of the model.

        :return: The coefficients.
        :rtype: torch.Tensor
        """
        return self._coefficients



"""Module for the Physics-Informed Neural Network solver."""

import torch

from .pinn_interface import PINNInterface
from ..solver import SingleSolverInterface
from ...problem import InverseProblem


class PINN(PINNInterface, SingleSolverInterface):
    r"""
    Physics-Informed Neural Network (PINN) solver class.
    This class implements Physics-Informed Neural Network solver, using a user
    specified ``model`` to solve a specific ``problem``.
    It can be used to solve both forward and inverse problems.

    The Physics Informed Neural Network solver aims to find the solution
    :math:`\mathbf{u}:\Omega\rightarrow\mathbb{R}^m` of a differential problem:

    .. math::

        \begin{cases}
        \mathcal{A}[\mathbf{u}](\mathbf{x})=0\quad,\mathbf{x}\in\Omega\\
        \mathcal{B}[\mathbf{u}](\mathbf{x})=0\quad,
        \mathbf{x}\in\partial\Omega
        \end{cases}

    minimizing the loss function:

    .. math::
        \mathcal{L}_{\rm{problem}} = \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(\mathcal{A}[\mathbf{u}](\mathbf{x}_i)) +
        \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(\mathcal{B}[\mathbf{u}](\mathbf{x}_i)),

    where :math:`\mathcal{L}` is a specific loss function, typically the MSE:

    .. math::
        \mathcal{L}(v) = \| v \|^2_2.

    .. seealso::

        **Original reference**: Karniadakis, G. E., Kevrekidis, I. G., Lu, L.,
        Perdikaris, P., Wang, S., & Yang, L. (2021).
        *Physics-informed machine learning.*
        Nature Reviews Physics, 3, 422-440.
        DOI: `10.1038 <https://doi.org/10.1038/s42254-021-00314-5>`_.
    """

    def __init__(
        self,
        problem,
        model,
        optimizer=None,
        scheduler=None,
        weighting=None,
        loss=None,
    ):
        """
        Initialization of the :class:`PINN` class.

        :param AbstractProblem problem: The problem to be solved.
        :param torch.nn.Module model: The neural network model to be used.
        :param Optimizer optimizer: The optimizer to be used.
            If ``None``, the :class:`torch.optim.Adam` optimizer is used.
            Default is ``None``.
        :param Scheduler scheduler: Learning rate scheduler.
            If ``None``, the :class:`torch.optim.lr_scheduler.ConstantLR`
            scheduler is used. Default is ``None``.
        :param WeightingInterface weighting: The weighting schema to be used.
            If ``None``, no weighting schema is used. Default is ``None``.
        :param torch.nn.Module loss: The loss function to be minimized.
            If ``None``, the :class:`torch.nn.MSELoss` loss is used.
            Default is `None`.
        """
        super().__init__(
            model=model,
            problem=problem,
            optimizer=optimizer,
            scheduler=scheduler,
            weighting=weighting,
            loss=loss,
        )

    def loss_data(self, input, target):
        """
        Compute the data loss for the PINN solver by evaluating the loss
        between the network's output and the true solution. This method should
        not be overridden, if not intentionally.

        :param input: The input to the neural network.
        :type input: LabelTensor
        :param target: The target to compare with the network's output.
        :type target: LabelTensor
        :return: The supervised loss, averaged over the number of observations.
        :rtype: LabelTensor
        """
        return self._loss_fn(self.forward(input), target)

    def loss_phys(self, samples, equation):
        """
        Computes the physics loss for the physics-informed solver based on the
        provided samples and equation.

        :param LabelTensor samples: The samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation.
        :return: The computed physics loss.
        :rtype: LabelTensor
        """
        residuals = self.compute_residual(samples, equation)
        return self._loss_fn(residuals, torch.zeros_like(residuals))

    def configure_optimizers(self):
        """
        Optimizer configuration for the PINN solver.

        :return: The optimizers and the schedulers
        :rtype: tuple[list[Optimizer], list[Scheduler]]
        """
        # If the problem is an InverseProblem, add the unknown parameters
        # to the parameters to be optimized.
        self.optimizer.hook(self.model.parameters())
        if isinstance(self.problem, InverseProblem):
            self.optimizer.instance.add_param_group(
                {
                    "params": [
                        self._params[var]
                        for var in self.problem.unknown_variables
                    ]
                }
            )
        self.scheduler.hook(self.optimizer)
        return ([self.optimizer.instance], [self.scheduler.instance])



"""Module for the LinearWeighting class."""

from ..loss import WeightingInterface
from ..utils import check_consistency, check_positive_integer


class LinearWeighting(WeightingInterface):
    """
    A weighting scheme that linearly scales weights from initial values to final
    values over a specified number of epochs.
    """

    def __init__(self, initial_weights, final_weights, target_epoch):
        """
        :param dict initial_weights: The weights to be assigned to each loss
            term at the beginning of training. The keys are the conditions and
            the values are the corresponding weights. If a condition is not
            present in the dictionary, the default value (1) is used.
        :param dict final_weights: The weights to be assigned to each loss term
            once the target epoch is reached. The keys are the conditions and
            the values are the corresponding weights. If a condition is not
            present in the dictionary, the default value (1) is used.
        :param int target_epoch: The epoch at which the weights reach their
            final values.
        :raises ValueError: If the keys of the two dictionaries are not
            consistent.
        """
        super().__init__(update_every_n_epochs=1, aggregator="sum")

        # Check consistency
        check_consistency([initial_weights, final_weights], dict)
        check_positive_integer(value=target_epoch, strict=True)

        # Check that the keys of the two dictionaries are the same
        if initial_weights.keys() != final_weights.keys():
            raise ValueError(
                "The keys of the initial_weights and final_weights "
                "dictionaries must be the same."
            )

        # Initialization
        self.initial_weights = initial_weights
        self.final_weights = final_weights
        self.target_epoch = target_epoch

    def weights_update(self, losses):
        """
        Update the weighting scheme based on the given losses.

        :param dict losses: The dictionary of losses.
        :return: The updated weights.
        :rtype: dict
        """
        return {
            condition: self.last_saved_weights().get(
                condition, self.initial_weights.get(condition, 1)
            )
            + (
                self.final_weights.get(condition, 1)
                - self.initial_weights.get(condition, 1)
            )
            / (self.target_epoch)
            for condition in losses.keys()
        }



"""Module for the InverseProblem class."""

from abc import abstractmethod
import torch
from .abstract_problem import AbstractProblem


class InverseProblem(AbstractProblem):
    """
    Class for defining inverse problems, where the objective is to determine
    unknown parameters through training, based on given data.
    """

    def __init__(self):
        """
        Initialization of the :class:`InverseProblem` class.
        """
        super().__init__()
        # storing unknown_parameters for optimization
        self.unknown_parameters = {}
        for var in self.unknown_variables:
            range_var = self.unknown_parameter_domain.range_[var]
            tensor_var = (
                torch.rand(1, requires_grad=True) * range_var[1] + range_var[0]
            )
            self.unknown_parameters[var] = torch.nn.Parameter(tensor_var)

    @abstractmethod
    def unknown_parameter_domain(self):
        """
        The domain of the unknown parameters of the problem.
        """

    @property
    def unknown_variables(self):
        """
        Get the unknown variables of the problem.

        :return: The unknown variables of the problem.
        :rtype: list[str]
        """
        return self.unknown_parameter_domain.variables

    @property
    def unknown_parameters(self):
        """
        Get the unknown parameters of the problem.

        :return: The unknown parameters of the problem.
        :rtype: torch.nn.Parameter
        """
        return self.__unknown_parameters

    @unknown_parameters.setter
    def unknown_parameters(self, value):
        """
        Set the unknown parameters of the problem.

        :param torch.nn.Parameter value: The unknown parameters of the problem.
        """
        self.__unknown_parameters = value



"""Module for the E(n) Equivariant Graph Neural Network block."""

import torch
from torch_geometric.nn import MessagePassing
from torch_geometric.utils import degree
from ....utils import check_positive_integer, check_consistency
from ....model import FeedForward


class EnEquivariantNetworkBlock(MessagePassing):
    """
    Implementation of the E(n) Equivariant Graph Neural Network block.
    This block is used to perform message-passing between nodes and edges in a
    graph neural network, following the scheme proposed by Satorras et al. in
    2021. It serves as an inner block in a larger graph neural network
    architecture.

    The message between two nodes connected by an edge is computed by applying a
    linear transformation to the sender node features and the edge features,
    together with the squared euclidean distance between the sender and
    recipient node positions, followed by a non-linear activation function.
    Messages are then aggregated using an aggregation scheme (e.g., sum, mean,
    min, max, or product).

    The update step is performed by applying another MLP to the concatenation of
    the incoming messages and the node features. Here, also the node
    positions are updated by adding the incoming messages divided by the
    degree of the recipient node.

    When velocity features are used, node velocities are passed through a small
    MLP to compute updates, which are then combined with the aggregated position
    messages. The node positions are updated both by the normalized position
    messages and by the updated velocities, ensuring equivariance while
    incorporating dynamic information.

    .. seealso::

        **Original reference** Satorras, V. G., Hoogeboom, E., Welling, M.
        (2021). *E(n) Equivariant Graph Neural Networks.*
        In International Conference on Machine Learning.
        DOI: `<https://doi.org/10.48550/arXiv.2102.09844>`_.
    """

    def __init__(
        self,
        node_feature_dim,
        edge_feature_dim,
        pos_dim,
        use_velocity=False,
        hidden_dim=64,
        n_message_layers=2,
        n_update_layers=2,
        activation=torch.nn.SiLU,
        aggr="add",
        node_dim=-2,
        flow="source_to_target",
    ):
        """
        Initialization of the :class:`EnEquivariantNetworkBlock` class.

        :param int node_feature_dim: The dimension of the node features.
        :param int edge_feature_dim: The dimension of the edge features.
        :param int pos_dim: The dimension of the position features.
        :param bool use_velocity: Whether to use velocity features in the
            message passing. Default is False.
        :param int hidden_dim: The dimension of the hidden features.
            Default is 64.
        :param int n_message_layers: The number of layers in the message
            network. Default is 2.
        :param int n_update_layers: The number of layers in the update network.
            Default is 2.
        :param torch.nn.Module activation: The activation function.
            Default is :class:`torch.nn.SiLU`.
        :param str aggr: The aggregation scheme to use for message passing.
            Available options are "add", "mean", "min", "max", "mul".
            See :class:`torch_geometric.nn.MessagePassing` for more details.
            Default is "add".
        :param int node_dim: The axis along which to propagate. Default is -2.
        :param str flow: The direction of message passing. Available options
            are "source_to_target" and "target_to_source".
            The "source_to_target" flow means that messages are sent from
            the source node to the target node, while the "target_to_source"
            flow means that messages are sent from the target node to the
            source node. See :class:`torch_geometric.nn.MessagePassing` for more
            details. Default is "source_to_target".
        :raises AssertionError: If `node_feature_dim` is not a positive integer.
        :raises AssertionError: If `edge_feature_dim` is a negative integer.
        :raises AssertionError: If `pos_dim` is not a positive integer.
        :raises AssertionError: If `hidden_dim` is not a positive integer.
        :raises AssertionError: If `n_message_layers` is not a positive integer.
        :raises AssertionError: If `n_update_layers` is not a positive integer.
        :raises AssertionError: If `use_velocity` is not a boolean.
        """
        super().__init__(aggr=aggr, node_dim=node_dim, flow=flow)

        # Check values
        check_positive_integer(node_feature_dim, strict=True)
        check_positive_integer(edge_feature_dim, strict=False)
        check_positive_integer(pos_dim, strict=True)
        check_positive_integer(hidden_dim, strict=True)
        check_positive_integer(n_message_layers, strict=True)
        check_positive_integer(n_update_layers, strict=True)
        check_consistency(use_velocity, bool)

        # Initialization
        self.use_velocity = use_velocity

        # Layer for computing the message
        self.message_net = FeedForward(
            input_dimensions=2 * node_feature_dim + edge_feature_dim + 1,
            output_dimensions=pos_dim,
            inner_size=hidden_dim,
            n_layers=n_message_layers,
            func=activation,
        )

        # Layer for updating the node features
        self.update_feat_net = FeedForward(
            input_dimensions=node_feature_dim + pos_dim,
            output_dimensions=node_feature_dim,
            inner_size=hidden_dim,
            n_layers=n_update_layers,
            func=activation,
        )

        # Layer for updating the node positions
        # The output dimension is set to 1 for equivariant updates
        self.update_pos_net = FeedForward(
            input_dimensions=pos_dim,
            output_dimensions=1,
            inner_size=hidden_dim,
            n_layers=n_update_layers,
            func=activation,
        )

        # If velocity is used, instantiate layer for velocity updates
        if self.use_velocity:
            self.update_vel_net = FeedForward(
                input_dimensions=node_feature_dim,
                output_dimensions=1,
                inner_size=hidden_dim,
                n_layers=n_update_layers,
                func=activation,
            )

    def forward(self, x, pos, edge_index, edge_attr=None, vel=None):
        """
        Forward pass of the block, triggering the message-passing routine.

        :param x: The node features.
        :type x: torch.Tensor | LabelTensor
        :param pos: The euclidean coordinates of the nodes.
        :type pos: torch.Tensor | LabelTensor
        :param torch.Tensor edge_index: The edge indices.
        :param edge_attr: The edge attributes. Default is None.
        :type edge_attr: torch.Tensor | LabelTensor
        :param vel: The velocity of the nodes. Default is None.
        :type vel: torch.Tensor | LabelTensor
        :return: The updated node features and node positions.
        :rtype: tuple(torch.Tensor, torch.Tensor)
        :raises: ValueError: If ``use_velocity`` is True and ``vel`` is None.
        """
        if self.use_velocity and vel is None:
            raise ValueError(
                "Velocity features are enabled, but no velocity is passed."
            )

        return self.propagate(
            edge_index=edge_index, x=x, pos=pos, edge_attr=edge_attr, vel=vel
        )

    def message(self, x_i, x_j, pos_i, pos_j, edge_attr):
        """
        Compute the message to be passed between nodes and edges.

        :param x_i: The node features of the recipient nodes.
        :type x_i: torch.Tensor | LabelTensor
        :param x_j: The node features of the sender nodes.
        :type x_j: torch.Tensor | LabelTensor
        :param pos_i: The node coordinates of the recipient nodes.
        :type pos_i: torch.Tensor | LabelTensor
        :param pos_j: The node coordinates of the sender nodes.
        :type pos_j: torch.Tensor | LabelTensor
        :param edge_attr: The edge attributes.
        :type edge_attr: torch.Tensor | LabelTensor
        :return: The message to be passed.
        :rtype: tuple(torch.Tensor, torch.Tensor)
        """
        # Compute the euclidean distance between the sender and recipient nodes
        diff = pos_i - pos_j
        dist = torch.norm(diff, dim=-1, keepdim=True) ** 2

        # Compute the message input
        if edge_attr is None:
            input_ = torch.cat((x_i, x_j, dist), dim=-1)
        else:
            input_ = torch.cat((x_i, x_j, dist, edge_attr), dim=-1)

        # Compute the messages and their equivariant counterpart
        m_ij = self.message_net(input_)
        message = diff * self.update_pos_net(m_ij)

        return message, m_ij

    def aggregate(self, inputs, index, ptr=None, dim_size=None):
        """
        Aggregate the messages at the nodes during message passing.

        This method receives a tuple of tensors corresponding to the messages
        to be aggregated. Both messages are aggregated separately according to
        the specified aggregation scheme.

        :param tuple(torch.Tensor) inputs: Tuple containing two messages to
            aggregate.
        :param index: The indices of target nodes for each message. This tensor
            specifies which node each message is aggregated into.
        :type index: torch.Tensor | LabelTensor
        :param ptr: Optional tensor to specify the slices of messages for each
            node (used in some aggregation strategies). Default is None.
        :type ptr: torch.Tensor | LabelTensor
        :param int dim_size: Optional size of the output dimension, i.e.,
            number of nodes. Default is None.
        :return: Tuple of aggregated tensors corresponding to (aggregated
            messages for position updates, aggregated messages for feature
            updates).
        :rtype: tuple(torch.Tensor, torch.Tensor)
        """
        # Unpack the messages from the inputs
        message, m_ij = inputs

        # Aggregate messages as usual using self.aggr method
        agg_message = super().aggregate(message, index, ptr, dim_size)
        agg_m_ij = super().aggregate(m_ij, index, ptr, dim_size)

        return agg_message, agg_m_ij

    def update(self, aggregated_inputs, x, pos, edge_index, vel):
        """
        Update node features, positions, and optionally velocities.

        :param tuple(torch.Tensor) aggregated_inputs: The messages to be passed.
        :param x: The node features.
        :type x: torch.Tensor | LabelTensor
        :param pos: The euclidean coordinates of the nodes.
        :type pos: torch.Tensor | LabelTensor
        :param torch.Tensor edge_index: The edge indices.
        :param vel: The velocity of the nodes.
        :type vel: torch.Tensor | LabelTensor
        :return: The updated node features and node positions.
        :rtype: tuple(torch.Tensor, torch.Tensor) |
            tuple(torch.Tensor, torch.Tensor, torch.Tensor)
        """
        # aggregated_inputs is tuple (agg_message, agg_m_ij)
        agg_message, agg_m_ij = aggregated_inputs

        # Degree for normalization of position updates
        c = degree(edge_index[1], pos.shape[0]).unsqueeze(-1).clamp(min=1)

        # If velocity is used, update it and use it to update positions
        if self.use_velocity:
            vel = self.update_vel_net(x) * vel

        # Update node features with aggregated m_ij
        x = self.update_feat_net(torch.cat((x, agg_m_ij), dim=-1))

        # Update positions with aggregated messages m_ij and velocities
        pos = pos + agg_message / c + (vel if self.use_velocity else 0)

        return (x, pos, vel) if self.use_velocity else (x, pos)



"""Modules for the the Embedding blocks."""

import torch
from pina.utils import check_consistency


class PeriodicBoundaryEmbedding(torch.nn.Module):
    r"""
    Enforcing hard-constrained periodic boundary conditions by embedding the
    input.

    A function :math:`u:\mathbb{R}^{\rm{in}} \rightarrow\mathbb{R}^{\rm{out}}`
    is periodic with respect to the spatial coordinates :math:`\mathbf{x}`
    with period :math:`\mathbf{L}` if:

    .. math::
        u(\mathbf{x})  = u(\mathbf{x} + n \mathbf{L})\;\;
        \forall n\in\mathbb{N}.

    The :class:`PeriodicBoundaryEmbedding` augments the input as follows:

    .. math::
        \mathbf{x} \rightarrow \tilde{\mathbf{x}} = \left[1,
        \cos\left(\frac{2\pi}{L_1} x_1 \right),
        \sin\left(\frac{2\pi}{L_1}x_1\right), \cdots,
        \cos\left(\frac{2\pi}{L_{\rm{in}}}x_{\rm{in}}\right),
        \sin\left(\frac{2\pi}{L_{\rm{in}}}x_{\rm{in}}\right)\right],

    where :math:`\text{dim}(\tilde{\mathbf{x}}) = 3\text{dim}(\mathbf{x})`.

    .. seealso::
        **Original reference**:
            1.  Dong, Suchuan, and Naxian Ni (2021).
                *A method for representing periodic functions and enforcing
                exactly periodic boundary conditions with deep neural networks*.
                Journal of Computational Physics 435, 110242.
                DOI: `10.1016/j.jcp.2021.110242.
                <https://doi.org/10.1016/j.jcp.2021.110242>`_
            2.  Wang, S., Sankaran, S., Wang, H., & Perdikaris, P. (2023).
                *An expert's guide to training physics-informed neural
                networks*.
                DOI: `arXiv preprint arXiv:2308.0846.
                <https://arxiv.org/abs/2308.08468>`_

    .. warning::
        The embedding is a truncated fourier expansion, and enforces periodic
        boundary conditions only for the function, and not for its derivatives.
        Enforcement of the approximate periodicity in the derivatives can be
        performed. Extensive tests have shown (see referenced papers) that this
        implementation can correctly enforce the periodic boundary conditions on
        the derivatives up to the order :math:`\sim 2,3`. This is not guaranteed
        for orders :math:`>3`. The PINA module is tested only for periodic
        boundary conditions on the function itself.
    """

    def __init__(self, input_dimension, periods, output_dimension=None):
        """
        Initialization of the :class:`PeriodicBoundaryEmbedding` block.

        :param int input_dimension: The dimension of the input tensor.
        :param periods: The periodicity with respect to each dimension for the
            input data. If ``float`` or ``int`` is passed, the period is assumed
            to be constant over all the dimensions of the data. If a ``dict`` is
            passed the `dict.values` represent periods, while the ``dict.keys``
            represent the dimension where the periodicity is enforced.
            The `dict.keys` can either be `int` if working with
            :class:`torch.Tensor`, or ``str`` if working with
            :class:`pina.label_tensor.LabelTensor`.
        :type periods: float | int | dict
        :param int output_dimension: The dimension of the output after the
            fourier embedding. If not ``None``, a :class:`torch.nn.Linear` layer
            is applied to the fourier embedding output to match the desired
            dimensionality. Default is ``None``.
        :raises TypeError: If the periods dict is not consistent.
        """
        super().__init__()

        # check input consistency
        check_consistency(periods, (float, int, dict))
        check_consistency(input_dimension, int)
        if output_dimension is not None:
            check_consistency(output_dimension, int)
            self._layer = torch.nn.Linear(input_dimension * 3, output_dimension)
        else:
            self._layer = torch.nn.Identity()

        # checks on the periods
        if isinstance(periods, dict):
            if not all(
                isinstance(dim, (str, int)) and isinstance(period, (float, int))
                for dim, period in periods.items()
            ):
                raise TypeError(
                    "In dictionary periods, keys must be integers"
                    " or strings, and values must be float or int."
                )
            self._period = periods
        else:
            self._period = {k: periods for k in range(input_dimension)}

    def forward(self, x):
        """
        Forward pass.

        :param x: The input tensor.
        :type x: torch.Tensor | LabelTensor
        :return: Periodic embedding of the input.
        :rtype: torch.Tensor
        """
        omega = torch.stack(
            [
                torch.pi * 2.0 / torch.tensor([val], device=x.device)
                for val in self._period.values()
            ],
            dim=-1,
        )
        x = self._get_vars(x, list(self._period.keys()))
        return self._layer(
            torch.cat(
                [
                    torch.ones_like(x),
                    torch.cos(omega * x),
                    torch.sin(omega * x),
                ],
                dim=-1,
            )
        )

    def _get_vars(self, x, indeces):
        """
        Get the variables from input tensor ordered by specific indeces.

        :param x: The input tensor from which to extract.
        :type x: torch.Tensor | LabelTensor
        :param indeces: The indeces to extract.
        :type indeces: list[int] | list[str]
        :raises RuntimeError: If the indeces are not consistent.
        :raises RuntimeError: If the extraction is not possible.
        :return: The extracted tensor.
        :rtype: torch.Tensor | LabelTensor
        """
        if isinstance(indeces[0], str):
            try:
                return x.extract(indeces)
            except AttributeError as e:
                raise RuntimeError(
                    "Not possible to extract input variables from tensor."
                    " Ensure that the passed tensor is a LabelTensor or"
                    " pass list of integers to extract variables. For"
                    " more information refer to warning in the documentation."
                ) from e
        elif isinstance(indeces[0], int):
            return x[..., indeces]
        else:
            raise RuntimeError(
                "Not able to extract correct indeces for tensor."
                " For more information refer to warning in the documentation."
            )

    @property
    def period(self):
        """
        The period of the function.

        :return: The period of the function.
        :rtype: dict | float | int
        """
        return self._period


class FourierFeatureEmbedding(torch.nn.Module):
    r"""
    Fourier Feature Embedding class to encode the input features using random
    Fourier features.

    This class applies a Fourier transformation to the input features, which can
    help in learning high-frequency variations in data. The class supports
    multiscale feature embedding, creating embeddings for each scale specified
    by the ``sigma`` parameter.

    The Fourier Feature Embedding augments the input features as follows
    (3.10 of original paper):

    .. math::
        \mathbf{x} \rightarrow \tilde{\mathbf{x}} = \left[
        \cos\left( \mathbf{B} \mathbf{x} \right),
        \sin\left( \mathbf{B} \mathbf{x} \right)\right],

    where :math:`\mathbf{B}_{ij} \sim \mathcal{N}(0, \sigma^2)`.

    If multiple ``sigma`` are passed, the resulting embeddings are concateneted:

    .. math::
        \mathbf{x} \rightarrow \tilde{\mathbf{x}} = \left[
        \cos\left( \mathbf{B}^1 \mathbf{x} \right),
        \sin\left( \mathbf{B}^1 \mathbf{x} \right),
        \cos\left( \mathbf{B}^2 \mathbf{x} \right),
        \sin\left( \mathbf{B}^3 \mathbf{x} \right),
        \dots,
        \cos\left( \mathbf{B}^M \mathbf{x} \right),
        \sin\left( \mathbf{B}^M \mathbf{x} \right)\right],

    where :math:`\mathbf{B}^k_{ij} \sim \mathcal{N}(0, \sigma_k^2) \quad k \in
    (1, \dots, M)`.

    .. seealso::
        **Original reference**:
        Wang, S., Wang, H., and Perdikaris, P. (2021).
        *On the eigenvector bias of Fourier feature networks: From regression to
        solving multi-scale PDEs with physics-informed neural networks.*
        Computer Methods in Applied Mechanics and Engineering 384 (2021):
        113938.
        DOI: `10.1016/j.cma.2021.113938.
        <https://doi.org/10.1016/j.cma.2021.113938>`_
    """

    def __init__(self, input_dimension, output_dimension, sigma):
        """
        Initialization of the :class:`FourierFeatureEmbedding` block.

        :param int input_dimension: The dimension of the input tensor.
        :param int output_dimension: The dimension of the output tensor. The
            output is obtained as a concatenation of cosine and sine embeddings.
        :param sigma: The standard deviation used for the Fourier Embedding.
            This value must reflect the granularity of the scale in the
            differential equation solution.
        :type sigma: float | int
        :raises RuntimeError: If the output dimension is not an even number.
        """
        super().__init__()

        # check consistency
        check_consistency(sigma, (int, float))
        check_consistency(output_dimension, int)
        check_consistency(input_dimension, int)
        if output_dimension % 2:
            raise RuntimeError(
                "Expected output_dimension to be a even number, "
                f"got {output_dimension}."
            )

        # assign sigma
        self._sigma = sigma

        # create non-trainable matrices
        self._matrix = (
            torch.rand(
                size=(input_dimension, output_dimension // 2),
                requires_grad=False,
            )
            * self.sigma
        )

    def forward(self, x):
        """
        Forward pass.

        :param x: The input tensor.
        :type x: torch.Tensor | LabelTensor
        :return: Fourier embedding of the input.
        :rtype: torch.Tensor
        """
        # compute random matrix multiplication
        out = torch.mm(x, self._matrix.to(device=x.device, dtype=x.dtype))
        # return embedding
        return torch.cat(
            [torch.cos(2 * torch.pi * out), torch.sin(2 * torch.pi * out)],
            dim=-1,
        )

    @property
    def sigma(self):
        """
        The standard deviation used for the Fourier Embedding.

        :return: The standard deviation used for the Fourier Embedding.
        :rtype: float | int
        """
        return self._sigma



"""Module for the Averaging Neural Operator Block class."""

import torch
from torch import nn
from ...utils import check_consistency


class AVNOBlock(nn.Module):
    r"""
    The inner block of the Averaging Neural Operator.

    The operator layer performs an affine transformation where the convolution
    is approximated with a local average. Given the input function
    :math:`v(x)\in\mathbb{R}^{\rm{emb}}` the layer computes the operator update
    :math:`K(v)` as:

    .. math::
        K(v) = \sigma\left(Wv(x) + b + \frac{1}{|\mathcal{A}|}\int v(y)dy\right)

    where:

    *   :math:`\mathbb{R}^{\rm{emb}}` is the embedding (hidden) size
        corresponding to the ``hidden_size`` object
    *   :math:`\sigma` is a non-linear activation, corresponding to the
        ``func`` object
    *   :math:`W\in\mathbb{R}^{\rm{emb}\times\rm{emb}}` is a tunable matrix.
    *   :math:`b\in\mathbb{R}^{\rm{emb}}` is a tunable bias.

    .. seealso::

        **Original reference**: Lanthaler S., Li, Z., Stuart, A. (2020).
        *The Nonlocal Neural Operator: Universal Approximation*.
        DOI: `arXiv preprint arXiv:2304.13221.
        <https://arxiv.org/abs/2304.13221>`_
    """

    def __init__(self, hidden_size=100, func=nn.GELU):
        """
        Initialization of the :class:`AVNOBlock` class.

        :param int hidden_size: The size of the hidden layer.
            Defaults is ``100``.
        :param func: The activation function.
            Default is :class:`torch.nn.GELU`.
        """
        super().__init__()

        # Check type consistency
        check_consistency(hidden_size, int)
        check_consistency(func, nn.Module, subclass=True)
        # Assignment
        self._nn = nn.Linear(hidden_size, hidden_size)
        self._func = func()

    def forward(self, x):
        r"""
        Forward pass of the block. It performs a sum of local average and an
        affine transformation of the field.

        :param torch.Tensor x: The input tensor for performing the computation.
        :return: The output tensor.
        :rtype: torch.Tensor
        """
        return self._func(self._nn(x) + torch.mean(x, dim=1, keepdim=True))



"""Module for the Problems."""

__all__ = [
    "AbstractProblem",
    "SpatialProblem",
    "TimeDependentProblem",
    "ParametricProblem",
    "InverseProblem",
]

from .abstract_problem import AbstractProblem
from .spatial_problem import SpatialProblem
from .time_dependent_problem import TimeDependentProblem
from .parametric_problem import ParametricProblem
from .inverse_problem import InverseProblem



"""Formulation of the diffusion-reaction problem."""

import torch
from ... import Condition
from ...equation import Equation, FixedValue, DiffusionReaction
from ...problem import SpatialProblem, TimeDependentProblem
from ...utils import check_consistency
from ...domain import CartesianDomain


def initial_condition(input_, output_):
    """
    Definition of the initial condition of the diffusion-reaction problem.

    :param LabelTensor input_: The input data of the problem.
    :param LabelTensor output_: The output data of the problem.
    :return: The residual of the initial condition.
    :rtype: LabelTensor
    """
    x = input_.extract("x")
    u_0 = (
        torch.sin(x)
        + (1 / 2) * torch.sin(2 * x)
        + (1 / 3) * torch.sin(3 * x)
        + (1 / 4) * torch.sin(4 * x)
        + (1 / 8) * torch.sin(8 * x)
    )
    return output_ - u_0


class DiffusionReactionProblem(TimeDependentProblem, SpatialProblem):
    r"""
    Implementation of the diffusion-reaction problem in the spatial interval
    :math:`[-\pi, \pi]` and temporal interval :math:`[0, 1]`.

    .. seealso::
        **Original reference**: Si, Chenhao, et al. *Complex Physics-Informed
        Neural Network.* arXiv preprint arXiv:2502.04917 (2025).
        DOI: `arXiv:2502.04917 <https://arxiv.org/abs/2502.04917>`_.

    :Example:
        >>> problem = DiffusionReactionProblem()
    """

    output_variables = ["u"]
    spatial_domain = CartesianDomain({"x": [-torch.pi, torch.pi]})
    temporal_domain = CartesianDomain({"t": [0, 1]})

    domains = {
        "D": CartesianDomain({"x": [-torch.pi, torch.pi], "t": [0, 1]}),
        "g1": CartesianDomain({"x": -torch.pi, "t": [0, 1]}),
        "g2": CartesianDomain({"x": torch.pi, "t": [0, 1]}),
        "t0": CartesianDomain({"x": [-torch.pi, torch.pi], "t": 0.0}),
    }

    conditions = {
        "g1": Condition(domain="g1", equation=FixedValue(0.0)),
        "g2": Condition(domain="g2", equation=FixedValue(0.0)),
        "t0": Condition(domain="t0", equation=Equation(initial_condition)),
    }

    def __init__(self, alpha=1e-4):
        """
        Initialization of the :class:`DiffusionReactionProblem`.

        :param alpha: The diffusion coefficient.
        :type alpha: float | int
        """
        super().__init__()
        check_consistency(alpha, (float, int))
        self.alpha = alpha

        def forcing_term(input_):
            """
            Implementation of the forcing term.
            """
            # Extract spatial and temporal variables
            spatial_d = [di for di in input_.labels if di != "t"]
            x = input_.extract(spatial_d)
            t = input_.extract("t")

            return torch.exp(-t) * (
                1.5 * torch.sin(2 * x)
                + (8 / 3) * torch.sin(3 * x)
                + (15 / 4) * torch.sin(4 * x)
                + (63 / 8) * torch.sin(8 * x)
            )

        self.conditions["D"] = Condition(
            domain="D",
            equation=DiffusionReaction(self.alpha, forcing_term),
        )

    def solution(self, pts):
        """
        Implementation of the analytical solution of the diffusion-reaction
        problem.

        :param LabelTensor pts: Points where the solution is evaluated.
        :return: The analytical solution of the diffusion-reaction problem.
        :rtype: LabelTensor
        """
        t = pts.extract("t")
        x = pts.extract("x")
        sol = torch.exp(-t) * (
            torch.sin(x)
            + (1 / 2) * torch.sin(2 * x)
            + (1 / 3) * torch.sin(3 * x)
            + (1 / 4) * torch.sin(4 * x)
            + (1 / 8) * torch.sin(8 * x)
        )
        sol.labels = self.output_variables
        return sol



"""Module for the Physics-Informed solvers."""

__all__ = [
    "PINNInterface",
    "PINN",
    "GradientPINN",
    "CausalPINN",
    "CompetitivePINN",
    "SelfAdaptivePINN",
    "RBAPINN",
]

from .pinn_interface import PINNInterface
from .pinn import PINN
from .rba_pinn import RBAPINN
from .causal_pinn import CausalPINN
from .gradient_pinn import GradientPINN
from .competitive_pinn import CompetitivePINN
from .self_adaptive_pinn import SelfAdaptivePINN



"""Module for the building blocks of the neural models."""

__all__ = [
    "ContinuousConvBlock",
    "ResidualBlock",
    "EnhancedLinear",
    "SpectralConvBlock1D",
    "SpectralConvBlock2D",
    "SpectralConvBlock3D",
    "FourierBlock1D",
    "FourierBlock2D",
    "FourierBlock3D",
    "PODBlock",
    "OrthogonalBlock",
    "PeriodicBoundaryEmbedding",
    "FourierFeatureEmbedding",
    "AVNOBlock",
    "LowRankBlock",
    "RBFBlock",
    "GNOBlock",
    "PirateNetBlock",
]

from .convolution_2d import ContinuousConvBlock
from .residual import ResidualBlock, EnhancedLinear
from .spectral import (
    SpectralConvBlock1D,
    SpectralConvBlock2D,
    SpectralConvBlock3D,
)
from .fourier_block import FourierBlock1D, FourierBlock2D, FourierBlock3D
from .pod_block import PODBlock
from .orthogonal import OrthogonalBlock
from .embedding import PeriodicBoundaryEmbedding, FourierFeatureEmbedding
from .average_neural_operator_block import AVNOBlock
from .low_rank_block import LowRankBlock
from .rbf_block import RBFBlock
from .gno_block import GNOBlock
from .pirate_network_block import PirateNetBlock



"""Module for the bivariate B-Spline surface model class."""

import torch
from .spline import Spline
from ..utils import check_consistency


class SplineSurface(torch.nn.Module):
    r"""
    The bivariate B-Spline surface model class.

    A bivariate B-spline surface is a parametric surface defined as the tensor
    product of two univariate B-spline curves:

    .. math::

        S(x, y) = \sum_{i,j=1}^{n_x, n_y} B_{i,k}(x) B_{j,s}(y) C_{i,j},
        \quad x \in [x_1, x_m], y \in [y_1, y_l]

    where:

    - :math:`C_{i,j} \in \mathbb{R}^2` are the control points. These fixed
      points influence the shape of the surface but are not generally
      interpolated, except at the boundaries under certain knot multiplicities.
    - :math:`B_{i,k}(x)` and :math:`B_{j,s}(y)` are the B-spline basis functions
      defined over two orthogonal directions, with orders :math:`k` and
      :math:`s`, respectively.
    - :math:`X = \{ x_1, x_2, \dots, x_m \}` and
      :math:`Y = \{ y_1, y_2, \dots, y_l \}` are the non-decreasing knot
      vectors along the two directions.
    """

    def __init__(self, orders, knots_u=None, knots_v=None, control_points=None):
        """
        Initialization of the :class:`SplineSurface` class.

        :param list[int] orders: The orders of the spline along each parametric
            direction. Each order defines the degree of the corresponding basis
            as ``degree = order - 1``.
        :param knots_u: The knots of the spline along the first direction.
            For details on valid formats and initialization modes, see the
            :class:`Spline` class. Default is None.
        :type knots_u: torch.Tensor | dict
        :param knots_v: The knots of the spline along the second direction.
            For details on valid formats and initialization modes, see the
            :class:`Spline` class. Default is None.
        :type knots_v: torch.Tensor | dict
        :param torch.Tensor control_points: The control points defining the
            surface geometry. It must be a two-dimensional tensor of shape
            ``[len(knots_u) - orders[0], len(knots_v) - orders[1]]``.
            If None, they are initialized as learnable parameters with zero
            values. Default is None.
        :raises ValueError: If ``orders`` is not a list of integers.
        :raises ValueError: If ``knots_u`` is neither a torch.Tensor nor a
            dictionary, when provided.
        :raises ValueError: If ``knots_v`` is neither a torch.Tensor nor a
            dictionary, when provided.
        :raises ValueError: If ``control_points`` is not a torch.Tensor,
            when provided.
        :raises ValueError: If ``orders`` is not a list of two elements.
        :raises ValueError: If ``knots_u``, ``knots_v``, and ``control_points``
            are all None.
        """
        super().__init__()

        # Check consistency
        check_consistency(orders, int)
        check_consistency(control_points, (type(None), torch.Tensor))
        check_consistency(knots_u, (type(None), torch.Tensor, dict))
        check_consistency(knots_v, (type(None), torch.Tensor, dict))

        # Check orders is a list of two elements
        if len(orders) != 2:
            raise ValueError("orders must be a list of two elements.")

        # Raise error if neither knots nor control points are provided
        if (knots_u is None or knots_v is None) and control_points is None:
            raise ValueError(
                "control_points cannot be None if knots_u or knots_v is None."
            )

        # Initialize knots_u if not provided
        if knots_u is None and control_points is not None:
            knots_u = {
                "n": control_points.shape[0] + orders[0],
                "min": 0,
                "max": 1,
                "mode": "auto",
            }

        # Initialize knots_v if not provided
        if knots_v is None and control_points is not None:
            knots_v = {
                "n": control_points.shape[1] + orders[1],
                "min": 0,
                "max": 1,
                "mode": "auto",
            }

        # Create two univariate b-splines
        self.spline_u = Spline(order=orders[0], knots=knots_u)
        self.spline_v = Spline(order=orders[1], knots=knots_v)
        self.control_points = control_points

        # Delete unneeded parameters
        delattr(self.spline_u, "_control_points")
        delattr(self.spline_v, "_control_points")

    def forward(self, x):
        """
        Forward pass for the :class:`SplineSurface` model.

        :param x: The input tensor.
        :type x: torch.Tensor | LabelTensor
        :return: The output tensor.
        :rtype: torch.Tensor
        """
        return torch.einsum(
            "...bi, ...bj, ij -> ...b",
            self.spline_u.basis(x.as_subclass(torch.Tensor)[..., 0]),
            self.spline_v.basis(x.as_subclass(torch.Tensor)[..., 1]),
            self.control_points,
        ).unsqueeze(-1)

    @property
    def knots(self):
        """
        The knots of the univariate splines defining the spline surface.

        :return: The knots.
        :rtype: tuple(torch.Tensor, torch.Tensor)
        """
        return self.spline_u.knots, self.spline_v.knots

    @knots.setter
    def knots(self, value):
        """
        Set the knots of the spline surface.

        :param value: A tuple (knots_u, knots_v) containing the knots for both
            parametric directions.
        :type value: tuple(torch.Tensor | dict, torch.Tensor | dict)
        :raises ValueError: If value is not a tuple of two elements.
        """
        # Check value is a tuple of two elements
        if not (isinstance(value, tuple) and len(value) == 2):
            raise ValueError("Knots must be a tuple of two elements.")

        knots_u, knots_v = value
        self.spline_u.knots = knots_u
        self.spline_v.knots = knots_v

    @property
    def control_points(self):
        """
        The control points of the spline.

        :return: The control points.
        :rtype: torch.Tensor
        """
        return self._control_points

    @control_points.setter
    def control_points(self, control_points):
        """
        Set the control points of the spline surface.

        :param torch.Tensor control_points: The bidimensional control points
            tensor, where each dimension refers to a direction in the parameter
            space. If None, control points are initialized to learnable
            parameters with zero initial value. Default is None.
        :raises ValueError: If in any direction there are not enough knots to
            define the control points, due to the relation:
            #knots = order + #control_points.
        :raises ValueError: If ``control_points`` is not of the correct shape.
        """
        # Save correct shape of control points
        __valid_shape = (
            len(self.spline_u.knots) - self.spline_u.order,
            len(self.spline_v.knots) - self.spline_v.order,
        )

        # If control points are not provided, initialize them
        if control_points is None:

            # Check that there are enough knots to define control points
            if (
                len(self.spline_u.knots) < self.spline_u.order + 1
                or len(self.spline_v.knots) < self.spline_v.order + 1
            ):
                raise ValueError(
                    f"Not enough knots to define control points. Got "
                    f"{len(self.spline_u.knots)} knots along u and "
                    f"{len(self.spline_v.knots)} knots along v, but need at "
                    f"least {self.spline_u.order + 1} and "
                    f"{self.spline_v.order + 1}, respectively."
                )

            # Initialize control points to zero
            control_points = torch.zeros(__valid_shape)

        # Check control points
        if control_points.shape != __valid_shape:
            raise ValueError(
                "control_points must be of the correct shape. ",
                f"Expected {__valid_shape}, got {control_points.shape}.",
            )

        # Register control points as a learnable parameter
        self._control_points = torch.nn.Parameter(
            control_points, requires_grad=True
        )



"""Module for the Simplex Domain."""

import torch
from .domain_interface import DomainInterface
from .cartesian import CartesianDomain
from ..label_tensor import LabelTensor
from ..utils import check_consistency


class SimplexDomain(DomainInterface):
    """
    Implementation of the simplex domain.
    """

    def __init__(self, simplex_matrix, sample_surface=False):
        """
        Initialization of the :class:`SimplexDomain` class.

        :param list[LabelTensor] simplex_matrix: A matrix representing the
            vertices of the simplex.
        :param bool sample_surface: A flag to choose the sampling strategy.
            If ``True``, samples are taken only from the surface of the simplex.
            If ``False``, samples are taken from the interior of the simplex.
            Default is ``False``.
        :raises ValueError: If the labels of the vertices don't match.
        :raises ValueError: If the number of vertices is not equal to the
            dimension of the simplex plus one.

        .. warning::
            Sampling for dimensions greater or equal to 10 could result in a
            shrinkage of the simplex, which degrades the quality of the samples.
            For dimensions higher than 10, use other sampling algorithms.

        :Example:
            >>> spatial_domain = SimplexDomain(
                    [
                        LabelTensor(torch.tensor([[0, 0]]), labels=["x", "y"]),
                        LabelTensor(torch.tensor([[1, 1]]), labels=["x", "y"]),
                        LabelTensor(torch.tensor([[0, 2]]), labels=["x", "y"]),
                    ], sample_surface = True
                )
        """

        # check consistency of sample_surface as bool
        check_consistency(sample_surface, bool)
        self._sample_surface = sample_surface

        # check consistency of simplex_matrix as list or tuple
        check_consistency([simplex_matrix], (list, tuple))

        # check everything within simplex_matrix is a LabelTensor
        check_consistency(simplex_matrix, LabelTensor)

        # check consistency of labels
        matrix_labels = simplex_matrix[0].labels
        if not all(vertex.labels == matrix_labels for vertex in simplex_matrix):
            raise ValueError("Labels don't match.")

        # check consistency dimensions
        dim_simplex = len(matrix_labels)
        if len(simplex_matrix) != dim_simplex + 1:
            raise ValueError(
                "An n-dimensional simplex is composed by n + 1 tensors of "
                "dimension n."
            )

        # creating vertices matrix
        self._vertices_matrix = LabelTensor.vstack(simplex_matrix)

        # creating basis vectors for simplex
        vert = self._vertices_matrix
        self._vectors_shifted = (vert[:-1] - vert[-1]).T

        # build cartesian_bound
        self._cartesian_bound = self._build_cartesian(self._vertices_matrix)

    @property
    def sample_modes(self):
        """
        List of available sampling modes.

        :return: List of available sampling modes.
        :rtype: list[str]
        """
        return ["random"]

    @property
    def variables(self):
        """
        List of variables of the domain.

        :return: List of variables of the domain.
        :rtype: list[str]
        """
        return sorted(self._vertices_matrix.labels)

    def _build_cartesian(self, vertices):
        """
        Build the cartesian border for a simplex domain to be used in sampling.

        :param list[LabelTensor] vertices: list of vertices defining the domain.
        :return: The cartesian border for the simplex domain.
        :rtype: CartesianDomain
        """

        span_dict = {}
        for coord in self.variables:
            sorted_vertices = torch.sort(vertices[coord].tensor.squeeze())
            # respective coord bounded by the lowest and highest values
            span_dict[coord] = [
                float(sorted_vertices.values[0]),
                float(sorted_vertices.values[-1]),
            ]

        return CartesianDomain(span_dict)

    def is_inside(self, point, check_border=False):
        """
        Check if a point is inside the simplex. It uses an algorithm involving
        barycentric coordinates.

        :param LabelTensor point: Point to be checked.
        :param check_border: If ``True``, the border is considered inside
            the simplex. Default is ``False``.
        :raises ValueError: If the labels of the point are different from those
            passed in the ``__init__`` method.
        :return: ``True`` if the point is inside the domain,
            ``False`` otherwise.
        :rtype: bool
        """

        if not all(label in self.variables for label in point.labels):
            raise ValueError(
                "Point labels different from constructor"
                f" dictionary labels. Got {point.labels},"
                f" expected {self.variables}."
            )

        point_shift = point - self._vertices_matrix[-1]
        point_shift = point_shift.tensor.reshape(-1, 1)

        # compute barycentric coordinates
        lambda_ = torch.linalg.solve(
            self._vectors_shifted * 1.0, point_shift * 1.0
        )
        lambda_1 = 1.0 - torch.sum(lambda_)
        lambdas = torch.vstack([lambda_, lambda_1])

        # perform checks
        if not check_border:
            return all(torch.gt(lambdas, 0.0)) and all(torch.lt(lambdas, 1.0))

        return all(torch.ge(lambdas, 0)) and (
            any(torch.eq(lambdas, 0)) or any(torch.eq(lambdas, 1))
        )

    def _sample_interior_randomly(self, n, variables):
        """
        Sample at random points from the interior of the simplex. Boundaries are
        excluded from this sampling routine.

        :param int n: Number of points to sample.
        :param list[str] variables: variables to be sampled.
        :return: Sampled points.
        :rtype: list[torch.Tensor]
        """

        # =============== For Developers ================ #
        #
        # The sampling startegy used is fairly simple.
        # First we sample a random vector from the hypercube
        # which contains the simplex. Then, if the point
        # sampled is inside the simplex, we add it as a valid
        # one.
        #
        # =============================================== #

        sampled_points = []
        while len(sampled_points) < n:
            sampled_point = self._cartesian_bound.sample(
                n=1, mode="random", variables=variables
            )

            if self.is_inside(sampled_point, self._sample_surface):
                sampled_points.append(sampled_point)
        return torch.cat(sampled_points, dim=0)

    def _sample_boundary_randomly(self, n):
        """
        Sample at random points from the boundary of the simplex.

        :param int n: Number of points to sample.
        :return: Sampled points.
        :rtype: torch.Tensor
        """

        # =============== For Developers ================ #
        #
        # The sampling startegy used is fairly simple.
        # We first sample the lambdas in [0, 1] domain,
        # we then set to zero only one lambda, and normalize.
        # Finally, we compute the matrix product between the
        # lamdas and the vertices matrix.
        #
        # =============================================== #

        sampled_points = []

        while len(sampled_points) < n:
            # extract number of vertices
            number_of_vertices = self._vertices_matrix.shape[0]
            # extract idx lambda to set to zero randomly
            idx_lambda = torch.randint(
                low=0, high=number_of_vertices, size=(1,)
            )
            # build lambda vector
            # 1. sampling [1, 2)
            lambdas = torch.rand((number_of_vertices, 1))
            # 2. setting lambdas[idx_lambda] to 0
            lambdas[idx_lambda] = 0
            # 3. normalize
            lambdas /= lambdas.sum()
            # 4. compute dot product
            sampled_points.append(self._vertices_matrix.T @ lambdas)
        return torch.cat(sampled_points, dim=1).T

    def sample(self, n, mode="random", variables="all"):
        """
        Sampling routine.

        :param int n: Number of points to sample.
        :param str mode: Sampling method. Default is ``random``.
            Available modes: random sampling, ``random``.
        :param list[str] variables: variables to be sampled. Default is ``all``.
        :raises NotImplementedError: If the sampling method is not implemented.
        :return: Sampled points.
        :rtype: LabelTensor

        .. warning::
            When ``sample_surface=True``, all variables are sampled,
            ignoring the ``variables`` parameter.
        """

        if variables == "all":
            variables = self.variables
        elif isinstance(variables, (list, tuple)):
            variables = sorted(variables)

        if mode in self.sample_modes:
            if self._sample_surface:
                sample_pts = self._sample_boundary_randomly(n)
            else:
                sample_pts = self._sample_interior_randomly(n, variables)

        else:
            raise NotImplementedError(f"mode={mode} is not implemented.")

        return LabelTensor(sample_pts, labels=self.variables)



"""Module for the DeepEnsemble solver interface."""

import torch
from ..solver import MultiSolverInterface
from ...utils import check_consistency


class DeepEnsembleSolverInterface(MultiSolverInterface):
    r"""
    A class for handling ensemble models in a multi-solver training framework.
    It allows for manual optimization, as well as the ability to train,
    validate, and test multiple models as part of an ensemble.
    The ensemble dimension can be customized to control how outputs are stacked.

    By default, it is compatible with problems defined by
    :class:`~pina.problem.abstract_problem.AbstractProblem`,
    and users can choose the problem type the solver is meant to address.

    An ensemble model is constructed by combining multiple models that solve
    the same type of problem. Mathematically, this creates an implicit
    distribution :math:`p(\mathbf{u} \mid \mathbf{s})` over the possible
    outputs :math:`\mathbf{u}`, given the original input :math:`\mathbf{s}`.
    The models :math:`\mathcal{M}_{i\in (1,\dots,r)}` in
    the ensemble work collaboratively to capture different
    aspects of the data or task, with each model contributing a distinct
    prediction :math:`\mathbf{y}_{i}=\mathcal{M}_i(\mathbf{u} \mid \mathbf{s})`.
    By aggregating these predictions, the ensemble
    model can achieve greater robustness and accuracy compared to individual
    models, leveraging the diversity of the models to reduce overfitting and
    improve generalization. Furthemore, statistical metrics can
    be computed, e.g. the ensemble mean and variance:

    .. math::
        \mathbf{\mu} = \frac{1}{N}\sum_{i=1}^r \mathbf{y}_{i}

    .. math::
        \mathbf{\sigma^2} = \frac{1}{N}\sum_{i=1}^r
        (\mathbf{y}_{i} - \mathbf{\mu})^2

    .. seealso::

        **Original reference**: Lakshminarayanan, B., Pritzel, A., & Blundell,
        C. (2017). *Simple and scalable predictive uncertainty estimation
        using deep ensembles*. Advances in neural information
        processing systems, 30.
        DOI: `arXiv:1612.01474 <https://arxiv.org/abs/1612.01474>`_.
    """

    def __init__(
        self,
        problem,
        models,
        optimizers=None,
        schedulers=None,
        weighting=None,
        use_lt=True,
        ensemble_dim=0,
    ):
        """
        Initialization of the :class:`DeepEnsembleSolverInterface` class.

        :param AbstractProblem problem: The problem to be solved.
        :param torch.nn.Module models: The neural network models to be used.
        :param Optimizer optimizer: The optimizer to be used.
            If ``None``, the :class:`torch.optim.Adam` optimizer is used.
            Default is ``None``.
        :param Scheduler scheduler: Learning rate scheduler.
            If ``None``, the :class:`torch.optim.lr_scheduler.ConstantLR`
            scheduler is used. Default is ``None``.
        :param WeightingInterface weighting: The weighting schema to be used.
            If ``None``, no weighting schema is used. Default is ``None``.
        :param bool use_lt: If ``True``, the solver uses LabelTensors as input.
            Default is ``True``.
        :param int ensemble_dim: The dimension along which the ensemble
            outputs are stacked. Default is 0.
        """
        super().__init__(
            problem, models, optimizers, schedulers, weighting, use_lt
        )
        # check consistency
        check_consistency(ensemble_dim, int)
        self._ensemble_dim = ensemble_dim

    def forward(self, x, ensemble_idx=None):
        """
        Forward pass through the ensemble models. If an `ensemble_idx` is
        provided, it returns the output of the specific model
        corresponding to that index. If no index is given, it stacks the outputs
        of all models along the ensemble dimension.

        :param LabelTensor x: The input tensor to the models.
        :param int ensemble_idx: Optional index to select a specific
            model from the ensemble. If ``None`` results for all models are
            stacked in ``ensemble_dim`` dimension. Default is ``None``.
        :return: The output of the selected model or the stacked
            outputs from all models.
        :rtype: LabelTensor
        """
        # if an index is passed, return the specific model output for that index
        if ensemble_idx is not None:
            return self.models[ensemble_idx].forward(x)
        # otherwise return the stacked output
        return torch.stack(
            [self.forward(x, idx) for idx in range(self.num_ensemble)],
            dim=self.ensemble_dim,
        )

    def training_step(self, batch):
        """
        Training step for the solver, overridden for manual optimization.
        This method performs a forward pass, calculates the loss, and applies
        manual backward propagation and optimization steps for each model in
        the ensemble.

        :param list[tuple[str, dict]] batch: A batch of training data.
            Each element is a tuple containing a condition name and a
            dictionary of points.
        :return: The aggregated loss after the training step.
        :rtype: torch.Tensor
        """
        # zero grad for optimizer
        for opt in self.optimizers:
            opt.instance.zero_grad()
        # perform forward passes and aggregate losses
        loss = super().training_step(batch)
        # perform backpropagation
        self.manual_backward(loss)
        # optimize
        for opt, sched in zip(self.optimizers, self.schedulers):
            opt.instance.step()
            sched.instance.step()
        return loss

    @property
    def ensemble_dim(self):
        """
        The dimension along which the ensemble outputs are stacked.

        :return: The ensemble dimension.
        :rtype: int
        """
        return self._ensemble_dim

    @property
    def num_ensemble(self):
        """
        The number of models in the ensemble.

        :return: The number of models in the ensemble.
        :rtype: int
        """
        return len(self.models)



"""Module for implemented problems."""

__all__ = [
    "SupervisedProblem",
    "HelmholtzProblem",
    "AllenCahnProblem",
    "AdvectionProblem",
    "Poisson2DSquareProblem",
    "DiffusionReactionProblem",
    "InversePoisson2DSquareProblem",
]

from .supervised_problem import SupervisedProblem
from .helmholtz import HelmholtzProblem
from .allen_cahn import AllenCahnProblem
from .advection import AdvectionProblem
from .poisson_2d_square import Poisson2DSquareProblem
from .diffusion_reaction import DiffusionReactionProblem
from .inverse_poisson_2d_square import InversePoisson2DSquareProblem



"""Module for the Equivariant Graph Neural Operator block."""

import torch
from ....utils import check_positive_integer
from .en_equivariant_network_block import EnEquivariantNetworkBlock


class EquivariantGraphNeuralOperatorBlock(torch.nn.Module):
    """
    A single block of the Equivariant Graph Neural Operator (EGNO).

    This block combines a temporal convolution with an equivariant graph neural
    network (EGNN) layer. It preserves equivariance while modeling complex
    interactions between nodes in a graph over time.

    .. seealso::

        **Original reference**
        Xu, M., Han, J., Lou, A., Kossaifi, J., Ramanathan, A., Azizzadenesheli,
        K., Leskovec, J., Ermon, S., Anandkumar, A. (2024).
        *Equivariant Graph Neural Operator for Modeling 3D Dynamics*
        DOI: `arXiv preprint arXiv:2401.11037.
        <https://arxiv.org/abs/2401.11037>`_
    """

    def __init__(
        self,
        node_feature_dim,
        edge_feature_dim,
        pos_dim,
        modes,
        hidden_dim=64,
        n_message_layers=2,
        n_update_layers=2,
        activation=torch.nn.SiLU,
        aggr="add",
        node_dim=-2,
        flow="source_to_target",
    ):
        """
        Initialization of the :class:`EquivariantGraphNeuralOperatorBlock`
        class.

        :param int node_feature_dim: The dimension of the node features.
        :param int edge_feature_dim: The dimension of the edge features.
        :param int pos_dim: The dimension of the position features.
        :param int modes: The number of Fourier modes to use in the temporal
            convolution.
        :param int hidden_dim: The dimension of the hidden features.
            Default is 64.
        :param int n_message_layers: The number of layers in the message
            network. Default is 2.
        :param int n_update_layers: The number of layers in the update network.
            Default is 2.
        :param torch.nn.Module activation: The activation function.
            Default is :class:`torch.nn.SiLU`.
        :param str aggr: The aggregation scheme to use for message passing.
            Available options are "add", "mean", "min", "max", "mul".
            See :class:`torch_geometric.nn.MessagePassing` for more details.
            Default is "add".
        :param int node_dim: The axis along which to propagate. Default is -2.
        :param str flow: The direction of message passing. Available options
            are "source_to_target" and "target_to_source".
            The "source_to_target" flow means that messages are sent from
            the source node to the target node, while the "target_to_source"
            flow means that messages are sent from the target node to the
            source node. See :class:`torch_geometric.nn.MessagePassing` for more
            details. Default is "source_to_target".
        :raises AssertionError: If ``modes`` is not a positive integer.
        """
        super().__init__()

        # Check consistency
        check_positive_integer(modes, strict=True)

        # Initialization
        self.modes = modes

        # Temporal convolution weights - real and imaginary parts
        self.weight_scalar_r = torch.nn.Parameter(
            torch.rand(node_feature_dim, node_feature_dim, modes)
        )
        self.weight_scalar_i = torch.nn.Parameter(
            torch.rand(node_feature_dim, node_feature_dim, modes)
        )
        self.weight_vector_r = torch.nn.Parameter(torch.rand(2, 2, modes) * 0.1)
        self.weight_vector_i = torch.nn.Parameter(torch.rand(2, 2, modes) * 0.1)

        # EGNN block
        self.egnn = EnEquivariantNetworkBlock(
            node_feature_dim=node_feature_dim,
            edge_feature_dim=edge_feature_dim,
            pos_dim=pos_dim,
            use_velocity=True,
            hidden_dim=hidden_dim,
            n_message_layers=n_message_layers,
            n_update_layers=n_update_layers,
            activation=activation,
            aggr=aggr,
            node_dim=node_dim,
            flow=flow,
        )

    def forward(self, x, pos, vel, edge_index, edge_attr=None):
        """
        Forward pass of the Equivariant Graph Neural Operator block.

        :param x: The node feature tensor of shape
            ``[time_steps, num_nodes, node_feature_dim]``.
        :type x: torch.Tensor | LabelTensor
        :param pos: The node position tensor (Euclidean coordinates) of shape
            ``[time_steps, num_nodes, pos_dim]``.
        :type pos: torch.Tensor | LabelTensor
        :param vel: The node velocity tensor of shape
            ``[time_steps, num_nodes, pos_dim]``.
        :type vel: torch.Tensor | LabelTensor
        :param edge_index: The edge connectivity of shape ``[2, num_edges]``.
        :type edge_index: torch.Tensor
        :param edge_attr: The edge feature tensor of shape
            ``[time_steps, num_edges, edge_feature_dim]``. Default is None.
        :type edge_attr: torch.Tensor | LabelTensor, optional
        :return: The updated node features, positions, and velocities, each with
            the same shape as the inputs.
        :rtype: tuple[torch.Tensor, torch.Tensor, torch.Tensor]
        """
        # Prepare features
        center = pos.mean(dim=1, keepdim=True)
        vector = torch.stack((pos - center, vel), dim=-1)

        # Compute temporal convolution
        x = x + self._convolution(
            x, "mni, iom -> mno", self.weight_scalar_r, self.weight_scalar_i
        )
        vector = vector + self._convolution(
            vector,
            "mndi, iom -> mndo",
            self.weight_vector_r,
            self.weight_vector_i,
        )

        # Split position and velocity
        pos, vel = vector.unbind(dim=-1)
        pos = pos + center

        # Reshape to (time * nodes, feature) for egnn
        x = x.reshape(-1, x.shape[-1])
        pos = pos.reshape(-1, pos.shape[-1])
        vel = vel.reshape(-1, vel.shape[-1])
        if edge_attr is not None:
            edge_attr = edge_attr.reshape(-1, edge_attr.shape[-1])

        x, pos, vel = self.egnn(
            x=x,
            pos=pos,
            edge_index=edge_index,
            edge_attr=edge_attr,
            vel=vel,
        )

        # Reshape back to (time, nodes, feature)
        x = x.reshape(center.shape[0], -1, x.shape[-1])
        pos = pos.reshape(center.shape[0], -1, pos.shape[-1])
        vel = vel.reshape(center.shape[0], -1, vel.shape[-1])

        return x, pos, vel

    def _convolution(self, x, einsum_idx, real, img):
        """
        Compute the temporal convolution.

        :param torch.Tensor x: The input features.
        :param str einsum_idx: The indices for the einsum operation.
        :param torch.Tensor real: The real part of the convolution weights.
        :param torch.Tensor img: The imaginary part of the convolution weights.
        :return: The convolved features.
        :rtype: torch.Tensor
        """
        # Number of modes to use
        modes = min(self.modes, (x.shape[0] // 2) + 1)

        # Build complex weights
        weights = torch.complex(real[..., :modes], img[..., :modes])

        # Convolution in Fourier space
        fourier = torch.fft.rfftn(x, dim=[0])[:modes]
        out = torch.einsum(einsum_idx, fourier, weights)

        return torch.fft.irfftn(out, s=x.shape[0], dim=0)



"""Module for the Scalar Weighting."""

from .weighting_interface import WeightingInterface
from ..utils import check_consistency


class ScalarWeighting(WeightingInterface):
    """
    Weighting scheme that assigns a scalar weight to each loss term.
    """

    def __init__(self, weights):
        """
        Initialization of the :class:`ScalarWeighting` class.

        :param weights: The weights to be assigned to each loss term.
            If a single scalar value is provided, it is assigned to all loss
            terms. If a dictionary is provided, the keys are the conditions and
            the values are the weights. If a condition is not present in the
            dictionary, the default value (1) is used.
        :type weights: float | int | dict
        """
        super().__init__(update_every_n_epochs=1, aggregator="sum")

        # Check consistency
        check_consistency([weights], (float, dict, int))

        # Initialization
        if isinstance(weights, dict):
            self.values = weights
            self.default_value_weights = 1
        else:
            self.values = {}
            self.default_value_weights = weights

    def weights_update(self, losses):
        """
        Update the weighting scheme based on the given losses.

        :param dict losses: The dictionary of losses.
        :return: The updated weights.
        :rtype: dict
        """
        return {
            condition: self.values.get(condition, self.default_value_weights)
            for condition in losses.keys()
        }


class _NoWeighting(ScalarWeighting):
    """
    Weighting scheme that does not apply any weighting to the losses.
    """

    def __init__(self):
        """
        Initialization of the :class:`_NoWeighting` class.
        """
        super().__init__(weights=1)



"""Module for the Condition interface."""

from abc import ABCMeta
from torch_geometric.data import Data
from ..label_tensor import LabelTensor
from ..graph import Graph


class ConditionInterface(metaclass=ABCMeta):
    """
    Abstract base class for PINA conditions. All specific conditions must
    inherit from this interface.

    Refer to :class:`pina.condition.condition.Condition` for a thorough
    description of all available conditions and how to instantiate them.
    """

    def __init__(self):
        """
        Initialization of the :class:`ConditionInterface` class.
        """
        self._problem = None

    @property
    def problem(self):
        """
        Return the problem associated with this condition.

        :return: Problem associated with this condition.
        :rtype: ~pina.problem.abstract_problem.AbstractProblem
        """
        return self._problem

    @problem.setter
    def problem(self, value):
        """
        Set the problem associated with this condition.

        :param pina.problem.abstract_problem.AbstractProblem value: The problem
            to associate with this condition
        """
        self._problem = value

    @staticmethod
    def _check_graph_list_consistency(data_list):
        """
        Check the consistency of the list of Data | Graph objects.
        The following checks are performed:

        - All elements in the list must be of the same type (either
          :class:`~torch_geometric.data.Data` or :class:`~pina.graph.Graph`).

        - All elements in the list must have the same keys.

        - The data type of each tensor must be consistent across all elements.

        - If a tensor is a :class:`~pina.label_tensor.LabelTensor`, its labels
          must also be consistent across all elements.

        :param data_list: The list of Data | Graph objects to check.
        :type data_list: list[Data] | list[Graph] | tuple[Data] | tuple[Graph]
        :raises ValueError: If the input types are invalid.
        :raises ValueError: If all elements in the list do not have the same
            keys.
        :raises ValueError: If the type of each tensor is not consistent across
            all elements in the list.
        :raises ValueError: If the labels of the LabelTensors are not consistent
            across all elements in the list.
        """
        # If the data is a Graph or Data object, perform no checks
        if isinstance(data_list, (Graph, Data)):
            return

        # Check all elements in the list are of the same type
        if not all(isinstance(i, (Graph, Data)) for i in data_list):
            raise ValueError(
                "Invalid input. Please, provide either Data or Graph objects."
            )

        # Store the keys, data types and labels of the first element
        data = data_list[0]
        keys = sorted(list(data.keys()))
        data_types = {name: tensor.__class__ for name, tensor in data.items()}
        labels = {
            name: tensor.labels
            for name, tensor in data.items()
            if isinstance(tensor, LabelTensor)
        }

        # Iterate over the list of Data | Graph objects
        for data in data_list[1:]:

            # Check that all elements in the list have the same keys
            if sorted(list(data.keys())) != keys:
                raise ValueError(
                    "All elements in the list must have the same keys."
                )

            # Iterate over the tensors in the current element
            for name, tensor in data.items():
                # Check that the type of each tensor is consistent
                if tensor.__class__ is not data_types[name]:
                    raise ValueError(
                        f"Data {name} must be a {data_types[name]}, got "
                        f"{tensor.__class__}"
                    )

                # Check that the labels of each LabelTensor are consistent
                if isinstance(tensor, LabelTensor):
                    if tensor.labels != labels[name]:
                        raise ValueError(
                            "LabelTensor must have the same labels"
                        )

    def __getattribute__(self, name):
        """
        Get an attribute from the object.

        :param str name: The name of the attribute to get.
        :return: The requested attribute.
        :rtype: Any
        """
        to_return = super().__getattribute__(name)
        if isinstance(to_return, (Graph, Data)):
            to_return = [to_return]
        return to_return



"""Module for the Equation."""

import inspect
from .equation_interface import EquationInterface


class Equation(EquationInterface):
    """
    Implementation of the Equation class. Every ``equation`` passed to a
    :class:`~pina.condition.condition.Condition` object must be either an
    instance of :class:`Equation` or
    :class:`~pina.equation.system_equation.SystemEquation`.
    """

    def __init__(self, equation):
        """
        Initialization of the :class:`Equation` class.

        :param Callable equation: A ``torch`` callable function used to compute
            the residual of a mathematical equation.
        :raises ValueError: If the equation is not a callable function.
        """
        if not callable(equation):
            raise ValueError(
                "equation must be a callable function."
                "Expected a callable function, got "
                f"{equation}"
            )
        # compute the signature
        sig = inspect.signature(equation)
        self.__len_sig = len(sig.parameters)
        self.__equation = equation

    def residual(self, input_, output_, params_=None):
        """
        Compute the residual of the equation.

        :param LabelTensor input_: Input points where the equation is evaluated.
        :param LabelTensor output_: Output tensor, eventually produced by a
            :class:`torch.nn.Module` instance.
        :param dict params_: Dictionary of unknown parameters, associated with a
            :class:`~pina.problem.inverse_problem.InverseProblem` instance.
            If the equation is not related to a
            :class:`~pina.problem.inverse_problem.InverseProblem` instance, the
            parameters must be initialized to ``None``. Default is ``None``.
        :return: The computed residual of the equation.
        :rtype: LabelTensor
        :raises RuntimeError: If the underlying equation signature length is not
            2 (direct problem) or 3 (inverse problem).
        """
        # Move the equation to the input_ device
        self.to(input_.device)

        # Call the underlying equation based on its signature length
        if self.__len_sig == 2:
            return self.__equation(input_, output_)
        if self.__len_sig == 3:
            return self.__equation(input_, output_, params_)
        raise RuntimeError(
            f"Unexpected number of arguments in equation: {self.__len_sig}. "
            "Expected either 2 (direct problem) or 3 (inverse problem)."
        )



import torch
import pytest

from pina import LabelTensor
from pina.utils import merge_tensors, check_consistency, check_positive_integer
from pina.domain import EllipsoidDomain, CartesianDomain, DomainInterface


def test_merge_tensors():
    tensor1 = LabelTensor(torch.rand((20, 3)), ["a", "b", "c"])
    tensor2 = LabelTensor(torch.zeros((20, 3)), ["d", "e", "f"])
    tensor3 = LabelTensor(torch.ones((30, 3)), ["g", "h", "i"])

    merged_tensor = merge_tensors((tensor1, tensor2, tensor3))
    assert tuple(merged_tensor.labels) == (
        "a",
        "b",
        "c",
        "d",
        "e",
        "f",
        "g",
        "h",
        "i",
    )
    assert merged_tensor.shape == (20 * 20 * 30, 9)
    assert torch.all(merged_tensor.extract(("d", "e", "f")) == 0)
    assert torch.all(merged_tensor.extract(("g", "h", "i")) == 1)


def test_check_consistency_correct():
    ellipsoid1 = EllipsoidDomain({"x": [1, 2], "y": [-2, 1]})
    example_input_pts = LabelTensor(torch.tensor([[0, 0, 0]]), ["x", "y", "z"])

    check_consistency(example_input_pts, torch.Tensor)
    check_consistency(CartesianDomain, DomainInterface, subclass=True)
    check_consistency(ellipsoid1, DomainInterface)


def test_check_consistency_incorrect():
    ellipsoid1 = EllipsoidDomain({"x": [1, 2], "y": [-2, 1]})
    example_input_pts = LabelTensor(torch.tensor([[0, 0, 0]]), ["x", "y", "z"])

    with pytest.raises(ValueError):
        check_consistency(example_input_pts, DomainInterface)
    with pytest.raises(ValueError):
        check_consistency(torch.Tensor, DomainInterface, subclass=True)
    with pytest.raises(ValueError):
        check_consistency(ellipsoid1, torch.Tensor)


@pytest.mark.parametrize("value", [0, 1, 2, 3, 10])
@pytest.mark.parametrize("strict", [True, False])
def test_check_positive_integer(value, strict):
    if value != 0:
        check_positive_integer(value, strict=strict)
    else:
        check_positive_integer(value, strict=False)

    # Should fail if value is negative
    with pytest.raises(AssertionError):
        check_positive_integer(-1, strict=strict)

    # Should fail if value is not an integer
    with pytest.raises(AssertionError):
        check_positive_integer(1.5, strict=strict)

    # Should fail if value is not a number
    with pytest.raises(AssertionError):
        check_positive_integer("string", strict=strict)



"""Module for the Difference Operation."""

import torch
from .operation_interface import OperationInterface
from ..label_tensor import LabelTensor


class Difference(OperationInterface):
    r"""
    Implementation of the difference operation between of a list of domains.

    Given two sets :math:`A` and :math:`B`, define the difference of the two
    sets as:

    .. math::
        A - B = \{x \mid x \in A \land x \not\in B\},

    where :math:`x` is a point in :math:`\mathbb{R}^N`.
    """

    def __init__(self, geometries):
        """
        Initialization of the :class:`Difference` class.

        :param list[DomainInterface] geometries: A list of instances of the
            :class:`~pina.domain.domain_interface.DomainInterface` class on
            which the difference operation is performed. The first domain in the
            list serves as the base from which points are sampled, while the
            remaining domains define the regions to be excluded from the base
            domain to compute the difference.

        :Example:
            >>> # Create two ellipsoid domains
            >>> ellipsoid1 = EllipsoidDomain({'x': [-1, 1], 'y': [-1, 1]})
            >>> ellipsoid2 = EllipsoidDomain({'x': [0, 2], 'y': [0, 2]})
            >>> # Define the difference between the domains
            >>> difference = Difference([ellipsoid1, ellipsoid2])
        """
        super().__init__(geometries)

    def is_inside(self, point, check_border=False):
        """
        Check if a point is inside the resulting domain.

        :param LabelTensor point: Point to be checked.
        :param bool check_border: If ``True``, the border is considered inside
            the domain. Default is ``False``.
        :return: ``True`` if the point is inside the domain,
            ``False`` otherwise.
        :rtype: bool
        """
        for geometry in self.geometries[1:]:
            if geometry.is_inside(point):
                return False
        return self.geometries[0].is_inside(point, check_border)

    def sample(self, n, mode="random", variables="all"):
        """
        Sampling routine.

        :param int n: Number of points to sample.
        :param str mode: Sampling method. Default is ``random``.
            Available modes: random sampling, ``random``;
        :param list[str] variables: variables to be sampled. Default is ``all``.
        :raises NotImplementedError: If the sampling method is not implemented.
        :return: Sampled points.
        :rtype: LabelTensor

        :Example:
            >>> # Create two Cartesian domains
            >>> cartesian1 = CartesianDomain({'x': [0, 2], 'y': [0, 2]})
            >>> cartesian2 = CartesianDomain({'x': [1, 3], 'y': [1, 3]})
            >>> # Define the difference between the domains
            >>> difference = Difference([cartesian1, cartesian2])
            >>> # Sampling
            >>> difference.sample(n=5)
                LabelTensor([[0.8400, 0.9179],
                            [0.9154, 0.5769],
                            [1.7403, 0.4835],
                            [0.9545, 1.2851],
                            [1.3726, 0.9831]])
            >>> len(difference.sample(n=5)
                5

        """
        if mode not in self.sample_modes:
            raise NotImplementedError(
                f"{mode} is not a valid mode for sampling."
            )

        sampled = []

        # sample the points
        while len(sampled) < n:
            # get sample point from first geometry
            point = self.geometries[0].sample(1, mode, variables)
            is_inside = False

            # check if point is inside any other geometry
            for geometry in self.geometries[1:]:
                # if point is inside any other geometry, break
                if geometry.is_inside(point):
                    is_inside = True
                    break
            # if point is not inside any other geometry, add to sampled
            if not is_inside:
                sampled.append(point)

        return LabelTensor(torch.cat(sampled), labels=self.variables)



"""Module for the PirateNet block class."""

import torch
from ...utils import check_consistency, check_positive_integer


class PirateNetBlock(torch.nn.Module):
    """
    The inner block of Physics-Informed residual adaptive network (PirateNet).

    The block consists of three dense layers with dual gating operations and an
    adaptive residual connection. The trainable ``alpha`` parameter controls
    the contribution of the residual connection.

    .. seealso::

        **Original reference**:
        Wang, S., Sankaran, S., Stinis., P., Perdikaris, P. (2025).
        *Simulating Three-dimensional Turbulence with Physics-informed Neural
        Networks*.
        DOI: `arXiv preprint arXiv:2507.08972.
        <https://arxiv.org/abs/2507.08972>`_
    """

    def __init__(self, inner_size, activation):
        """
        Initialization of the :class:`PirateNetBlock` class.

        :param int inner_size: The number of hidden units in the dense layers.
        :param torch.nn.Module activation: The activation function.
        """
        super().__init__()

        # Check consistency
        check_consistency(activation, torch.nn.Module, subclass=True)
        check_positive_integer(inner_size, strict=True)

        # Initialize the linear transformations of the dense layers
        self.linear1 = torch.nn.Linear(inner_size, inner_size)
        self.linear2 = torch.nn.Linear(inner_size, inner_size)
        self.linear3 = torch.nn.Linear(inner_size, inner_size)

        # Initialize the scales of the dense layers
        self.scale1 = torch.nn.Parameter(torch.zeros(inner_size))
        self.scale2 = torch.nn.Parameter(torch.zeros(inner_size))
        self.scale3 = torch.nn.Parameter(torch.zeros(inner_size))

        # Initialize the adaptive residual connection parameter
        self._alpha = torch.nn.Parameter(torch.zeros(1))

        # Initialize the activation function
        self.activation = activation()

    def forward(self, x, U, V):
        """
        Forward pass of the PirateNet block. It computes the output of the block
        by applying the dense layers with scaling, and combines the results with
        the input using the adaptive residual connection.

        :param x: The input tensor.
        :type x: torch.Tensor | LabelTensor
        :param torch.Tensor U: The first shared gating tensor. It must have the
            same shape as ``x``.
        :param torch.Tensor V: The second shared gating tensor. It must have the
            same shape as ``x``.
        :return: The output tensor of the block.
        :rtype: torch.Tensor | LabelTensor
        """
        # Compute the output of the first dense layer with scaling
        f = self.activation(self.linear1(x) * torch.exp(self.scale1))
        z1 = f * U + (1 - f) * V

        # Compute the output of the second dense layer with scaling
        g = self.activation(self.linear2(z1) * torch.exp(self.scale2))
        z2 = g * U + (1 - g) * V

        # Compute the output of the block
        h = self.activation(self.linear3(z2) * torch.exp(self.scale3))
        return self._alpha * h + (1 - self._alpha) * x

    @property
    def alpha(self):
        """
        Return the alpha parameter.

        :return: The alpha parameter controlling the residual connection.
        :rtype: torch.nn.Parameter
        """
        return self._alpha



"""
Module for Pina Refinement callbacks.
"""

__all__ = [
    "RefinementInterface",
    "R3Refinement",
]

from .refinement_interface import RefinementInterface
from .r3_refinement import R3Refinement



"""Solver module."""

from abc import ABCMeta, abstractmethod
import lightning
import torch

from torch._dynamo import OptimizedModule
from ..problem import AbstractProblem, InverseProblem
from ..optim import Optimizer, Scheduler, TorchOptimizer, TorchScheduler
from ..loss import WeightingInterface
from ..loss.scalar_weighting import _NoWeighting
from ..utils import check_consistency, labelize_forward


class SolverInterface(lightning.pytorch.LightningModule, metaclass=ABCMeta):
    """
    Abstract base class for PINA solvers. All specific solvers must inherit
    from this interface. This class extends
    :class:`~lightning.pytorch.core.LightningModule`, providing additional
    functionalities for defining and optimizing Deep Learning models.

    By inheriting from this base class, solvers gain access to built-in training
    loops, logging utilities, and optimization techniques.
    """

    def __init__(self, problem, weighting, use_lt):
        """
        Initialization of the :class:`SolverInterface` class.

        :param AbstractProblem problem: The problem to be solved.
        :param WeightingInterface weighting: The weighting schema to be used.
            If ``None``, no weighting schema is used. Default is ``None``.
        :param bool use_lt: If ``True``, the solver uses LabelTensors as input.
        """
        super().__init__()

        # check consistency of the problem
        check_consistency(problem, AbstractProblem)
        self._check_solver_consistency(problem)
        self._pina_problem = problem

        # check consistency of the weighting and hook the condition names
        if weighting is None:
            weighting = _NoWeighting()
        check_consistency(weighting, WeightingInterface)
        self._pina_weighting = weighting
        weighting._solver = self

        # check consistency use_lt
        check_consistency(use_lt, bool)
        self._use_lt = use_lt

        # if use_lt is true add extract operation in input
        if use_lt is True:
            self.forward = labelize_forward(
                forward=self.forward,
                input_variables=problem.input_variables,
                output_variables=problem.output_variables,
            )

        # PINA private attributes (some are overridden by derived classes)
        self._pina_problem = problem
        self._pina_models = None
        self._pina_optimizers = None
        self._pina_schedulers = None

        # inverse problem handling
        if isinstance(self.problem, InverseProblem):
            self._params = self.problem.unknown_parameters
            self._clamp_params = self._clamp_inverse_problem_params
        else:
            self._params = None
            self._clamp_params = lambda: None

    @abstractmethod
    def forward(self, *args, **kwargs):
        """
        Abstract method for the forward pass implementation.

        :param args: The input tensor.
        :type args: torch.Tensor | LabelTensor | Data | Graph
        :param dict kwargs: Additional keyword arguments.
        """

    @abstractmethod
    def optimization_cycle(self, batch):
        """
        The optimization cycle for the solvers.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :return: The losses computed for all conditions in the batch, casted
            to a subclass of :class:`torch.Tensor`. It should return a dict
            containing the condition name and the associated scalar loss.
        :rtype: dict
        """

    def training_step(self, batch, **kwargs):
        """
        Solver training step. It computes the optimization cycle and aggregates
        the losses using the ``weighting`` attribute.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :param dict kwargs: Additional keyword arguments passed to
            ``optimization_cycle``.
        :return: The loss of the training step.
        :rtype: torch.Tensor
        """
        loss = self._optimization_cycle(batch=batch, **kwargs)
        self.store_log("train_loss", loss, self.get_batch_size(batch))
        return loss

    def validation_step(self, batch, **kwargs):
        """
        Solver validation step. It computes the optimization cycle and
        averages the losses. No aggregation using the ``weighting`` attribute is
        performed.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :param dict kwargs: Additional keyword arguments passed to
            ``optimization_cycle``.
        :return: The loss of the training step.
        :rtype: torch.Tensor
        """
        losses = self.optimization_cycle(batch=batch, **kwargs)
        loss = (sum(losses.values()) / len(losses)).as_subclass(torch.Tensor)
        self.store_log("val_loss", loss, self.get_batch_size(batch))
        return loss

    def test_step(self, batch, **kwargs):
        """
        Solver test step. It computes the optimization cycle and
        averages the losses. No aggregation using the ``weighting`` attribute is
        performed.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :param dict kwargs: Additional keyword arguments passed to
            ``optimization_cycle``.
        :return: The loss of the training step.
        :rtype: torch.Tensor
        """
        losses = self.optimization_cycle(batch=batch, **kwargs)
        loss = (sum(losses.values()) / len(losses)).as_subclass(torch.Tensor)
        self.store_log("test_loss", loss, self.get_batch_size(batch))
        return loss

    def store_log(self, name, value, batch_size):
        """
        Store the log of the solver.

        :param str name: The name of the log.
        :param torch.Tensor value: The value of the log.
        :param int batch_size: The size of the batch.
        """

        self.log(
            name=name,
            value=value,
            batch_size=batch_size,
            **self.trainer.logging_kwargs,
        )

    def setup(self, stage):
        """
        This method is called at the start of the train and test process to
        compile the model if the :class:`~pina.trainer.Trainer`
        ``compile`` is ``True``.

        :param str stage: The current stage of the training process
            (e.g., ``fit``, ``validate``, ``test``, ``predict``).
        :return: The result of the parent class ``setup`` method.
        :rtype: Any
        """
        if self.trainer.compile and not self._is_compiled():
            self._setup_compile()
        return super().setup(stage)

    def _is_compiled(self):
        """
        Check if the model is compiled.

        :return: ``True`` if the model is compiled, ``False`` otherwise.
        :rtype: bool
        """
        for model in self._pina_models:
            if not isinstance(model, OptimizedModule):
                return False
        return True

    def _setup_compile(self):
        """
        Compile all models in the solver using ``torch.compile``.

        This method iterates through each model stored in the solver
        list and attempts to compile them for optimized execution. It supports
        models of type `torch.nn.Module` and `torch.nn.ModuleDict`. For models
        stored in a `ModuleDict`, each submodule is compiled individually.
        Models on Apple Silicon (MPS) use the 'eager' backend,
        while others use 'inductor'.

        :raises RuntimeError: If a model is neither `torch.nn.Module`
            nor `torch.nn.ModuleDict`.
        """
        for i, model in enumerate(self._pina_models):
            if isinstance(model, torch.nn.ModuleDict):
                for name, module in model.items():
                    self._pina_models[i][name] = self._compile_modules(module)
            elif isinstance(model, torch.nn.Module):
                self._pina_models[i] = self._compile_modules(model)
            else:
                raise RuntimeError(
                    "Compilation available only for "
                    "torch.nn.Module or torch.nn.ModuleDict."
                )

    def _check_solver_consistency(self, problem):
        """
        Check the consistency of the solver with the problem formulation.

        :param AbstractProblem problem: The problem to be solved.
        """
        for condition in problem.conditions.values():
            check_consistency(condition, self.accepted_conditions_types)

    def _optimization_cycle(self, batch, **kwargs):
        """
        Aggregate the loss for each condition in the batch.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :param dict kwargs: Additional keyword arguments passed to
            ``optimization_cycle``.
        :return: The losses computed for all conditions in the batch, casted
            to a subclass of :class:`torch.Tensor`. It should return a dict
            containing the condition name and the associated scalar loss.
        :rtype: dict
        """
        # compute losses
        losses = self.optimization_cycle(batch)
        # clamp unknown parameters in InverseProblem (if needed)
        self._clamp_params()
        # store log
        for name, value in losses.items():
            self.store_log(
                f"{name}_loss", value.item(), self.get_batch_size(batch)
            )
        # aggregate
        loss = self.weighting.aggregate(losses).as_subclass(torch.Tensor)
        return loss

    def _clamp_inverse_problem_params(self):
        """
        Clamps the parameters of the inverse problem solver to specified ranges.
        """
        for v in self._params:
            self._params[v].data.clamp_(
                self.problem.unknown_parameter_domain.range_[v][0],
                self.problem.unknown_parameter_domain.range_[v][1],
            )

    @staticmethod
    def _compile_modules(model):
        """
        Perform the compilation of the model.

        This method attempts to compile the given PyTorch model
        using ``torch.compile`` to improve execution performance. The
        backend is selected based on the device on which the model resides:
        ``eager`` is used for MPS devices (Apple Silicon), and ``inductor``
        is used for all others.

        If compilation fails, the method prints the error and returns the
        original, uncompiled model.

        :param torch.nn.Module model: The model to compile.
        :raises Exception: If the compilation fails.
        :return: The compiled model.
        :rtype: torch.nn.Module
        """
        model_device = next(model.parameters()).device
        try:
            if model_device == torch.device("mps:0"):
                model = torch.compile(model, backend="eager")
            else:
                model = torch.compile(model, backend="inductor")
        except Exception as e:
            print("Compilation failed, running in normal mode.:\n", e)
        return model

    @staticmethod
    def get_batch_size(batch):
        """
        Get the batch size.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :return: The size of the batch.
        :rtype: int
        """

        batch_size = 0
        for data in batch:
            batch_size += len(data[1]["input"])
        return batch_size

    @staticmethod
    def default_torch_optimizer():
        """
        Set the default optimizer to :class:`torch.optim.Adam`.

        :return: The default optimizer.
        :rtype: Optimizer
        """
        return TorchOptimizer(torch.optim.Adam, lr=0.001)

    @staticmethod
    def default_torch_scheduler():
        """
        Set the default scheduler to
        :class:`torch.optim.lr_scheduler.ConstantLR`.

        :return: The default scheduler.
        :rtype: Scheduler
        """

        return TorchScheduler(torch.optim.lr_scheduler.ConstantLR)

    @property
    def problem(self):
        """
        The problem instance.

        :return: The problem instance.
        :rtype: :class:`~pina.problem.abstract_problem.AbstractProblem`
        """
        return self._pina_problem

    @property
    def use_lt(self):
        """
        Using LabelTensors as input during training.

        :return: The use_lt attribute.
        :rtype: bool
        """
        return self._use_lt

    @property
    def weighting(self):
        """
        The weighting schema.

        :return: The weighting schema.
        :rtype: :class:`~pina.loss.weighting_interface.WeightingInterface`
        """
        return self._pina_weighting


class SingleSolverInterface(SolverInterface, metaclass=ABCMeta):
    """
    Base class for PINA solvers using a single :class:`torch.nn.Module`.
    """

    def __init__(
        self,
        problem,
        model,
        optimizer=None,
        scheduler=None,
        weighting=None,
        use_lt=True,
    ):
        """
        Initialization of the :class:`SingleSolverInterface` class.

        :param AbstractProblem problem: The problem to be solved.
        :param torch.nn.Module model: The neural network model to be used.
        :param Optimizer optimizer: The optimizer to be used.
            If ``None``, the :class:`torch.optim.Adam` optimizer is
            used. Default is ``None``.
        :param Scheduler scheduler: The scheduler to be used.
            If ``None``, the :class:`torch.optim.lr_scheduler.ConstantLR`
            scheduler is used. Default is ``None``.
        :param WeightingInterface weighting: The weighting schema to be used.
            If ``None``, no weighting schema is used. Default is ``None``.
        :param bool use_lt: If ``True``, the solver uses LabelTensors as input.
        """
        if optimizer is None:
            optimizer = self.default_torch_optimizer()

        if scheduler is None:
            scheduler = self.default_torch_scheduler()

        super().__init__(problem=problem, use_lt=use_lt, weighting=weighting)

        # check consistency of models argument and encapsulate in list
        check_consistency(model, torch.nn.Module)
        # check scheduler consistency and encapsulate in list
        check_consistency(scheduler, Scheduler)
        # check optimizer consistency and encapsulate in list
        check_consistency(optimizer, Optimizer)

        # initialize the model (needed by Lightining to go to different devices)
        self._pina_models = torch.nn.ModuleList([model])
        self._pina_optimizers = [optimizer]
        self._pina_schedulers = [scheduler]

    def forward(self, x):
        """
        Forward pass implementation.

        :param x: Input tensor.
        :type x: torch.Tensor | LabelTensor | Graph | Data
        :return: Solver solution.
        :rtype: torch.Tensor | LabelTensor | Graph | Data
        """
        return self.model(x)

    def configure_optimizers(self):
        """
        Optimizer configuration for the solver.

        :return: The optimizer and the scheduler
        :rtype: tuple[list[Optimizer], list[Scheduler]]
        """
        self.optimizer.hook(self.model.parameters())
        if isinstance(self.problem, InverseProblem):
            self.optimizer.instance.add_param_group(
                {
                    "params": [
                        self._params[var]
                        for var in self.problem.unknown_variables
                    ]
                }
            )
        self.scheduler.hook(self.optimizer)
        return ([self.optimizer.instance], [self.scheduler.instance])

    @property
    def model(self):
        """
        The model used for training.

        :return: The model used for training.
        :rtype: torch.nn.Module
        """
        return self._pina_models[0]

    @property
    def scheduler(self):
        """
        The scheduler used for training.

        :return: The scheduler used for training.
        :rtype: Scheduler
        """
        return self._pina_schedulers[0]

    @property
    def optimizer(self):
        """
        The optimizer used for training.

        :return: The optimizer used for training.
        :rtype: Optimizer
        """
        return self._pina_optimizers[0]


class MultiSolverInterface(SolverInterface, metaclass=ABCMeta):
    """
    Base class for PINA solvers using multiple :class:`torch.nn.Module`.
    """

    def __init__(
        self,
        problem,
        models,
        optimizers=None,
        schedulers=None,
        weighting=None,
        use_lt=True,
    ):
        """
        Initialization of the :class:`MultiSolverInterface` class.

        :param AbstractProblem problem: The problem to be solved.
        :param models: The neural network models to be used.
        :type model: list[torch.nn.Module] | tuple[torch.nn.Module]
        :param list[Optimizer] optimizers: The optimizers to be used.
            If ``None``, the :class:`torch.optim.Adam` optimizer is used for all
            models. Default is ``None``.
        :param list[Scheduler] schedulers: The schedulers to be used.
            If ``None``, the :class:`torch.optim.lr_scheduler.ConstantLR`
            scheduler is used for all the models. Default is ``None``.
        :param WeightingInterface weighting: The weighting schema to be used.
            If ``None``, no weighting schema is used. Default is ``None``.
        :param bool use_lt: If ``True``, the solver uses LabelTensors as input.
        :raises ValueError: If the models are not a list or tuple with length
            greater than one.

        .. warning::
            :class:`MultiSolverInterface` uses manual optimization by setting
            ``automatic_optimization=False`` in
            :class:`~lightning.pytorch.core.LightningModule`. For more
            information on manual optimization please
            see `here <https://lightning.ai/docs/pytorch/stable/\
                model/manual_optimization.html>`_.
        """
        if not isinstance(models, (list, tuple)) or len(models) < 2:
            raise ValueError(
                "models should be list[torch.nn.Module] or "
                "tuple[torch.nn.Module] with len greater than "
                "one."
            )

        if optimizers is None:
            optimizers = [
                self.default_torch_optimizer() for _ in range(len(models))
            ]

        if schedulers is None:
            schedulers = [
                self.default_torch_scheduler() for _ in range(len(models))
            ]

        if any(opt is None for opt in optimizers):
            optimizers = [
                self.default_torch_optimizer() if opt is None else opt
                for opt in optimizers
            ]

        if any(sched is None for sched in schedulers):
            schedulers = [
                self.default_torch_scheduler() if sched is None else sched
                for sched in schedulers
            ]

        super().__init__(problem=problem, use_lt=use_lt, weighting=weighting)

        # check consistency of models argument and encapsulate in list
        check_consistency(models, torch.nn.Module)

        # check scheduler consistency and encapsulate in list
        check_consistency(schedulers, Scheduler)

        # check optimizer consistency and encapsulate in list
        check_consistency(optimizers, Optimizer)

        # check length consistency optimizers
        if len(models) != len(optimizers):
            raise ValueError(
                "You must define one optimizer for each model."
                f"Got {len(models)} models, and {len(optimizers)}"
                " optimizers."
            )
        if len(schedulers) != len(optimizers):
            raise ValueError(
                "You must define one scheduler for each optimizer."
                f"Got {len(schedulers)} schedulers, and {len(optimizers)}"
                " optimizers."
            )

        # initialize the model
        self._pina_models = torch.nn.ModuleList(models)
        self._pina_optimizers = optimizers
        self._pina_schedulers = schedulers

        # Set automatic optimization to False.
        # For more information on manual optimization see:
        # http://lightning.ai/docs/pytorch/stable/model/manual_optimization.html
        self.automatic_optimization = False

    def on_train_batch_end(self, outputs, batch, batch_idx):
        """
        This method is called at the end of each training batch and overrides
        the PyTorch Lightning implementation to log checkpoints.

        :param torch.Tensor outputs: The ``model``'s output for the current
            batch.
        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :param int batch_idx: The index of the current batch.
        """
        # increase by one the counter of optimization to save loggers
        epoch_loop = self.trainer.fit_loop.epoch_loop
        epoch_loop.manual_optimization.optim_step_progress.total.completed += 1
        return super().on_train_batch_end(outputs, batch, batch_idx)

    def configure_optimizers(self):
        """
        Optimizer configuration for the solver.

        :return: The optimizer and the scheduler
        :rtype: tuple[list[Optimizer], list[Scheduler]]
        """
        for optimizer, scheduler, model in zip(
            self.optimizers, self.schedulers, self.models
        ):
            optimizer.hook(model.parameters())
            scheduler.hook(optimizer)

        return (
            [optimizer.instance for optimizer in self.optimizers],
            [scheduler.instance for scheduler in self.schedulers],
        )

    @property
    def models(self):
        """
        The models used for training.

        :return: The models used for training.
        :rtype: torch.nn.ModuleList
        """
        return self._pina_models

    @property
    def optimizers(self):
        """
        The optimizers used for training.

        :return: The optimizers used for training.
        :rtype: list[Optimizer]
        """
        return self._pina_optimizers

    @property
    def schedulers(self):
        """
        The schedulers used for training.

        :return: The schedulers used for training.
        :rtype: list[Scheduler]
        """
        return self._pina_schedulers



"""Module for the Residual-Based Attention PINN solver."""

import torch

from .pinn import PINN
from ...utils import check_consistency


class RBAPINN(PINN):
    r"""
    Residual-based Attention Physics-Informed Neural Network (RBAPINN) solver
    class. This class implements the Residual-based Attention Physics-Informed
    Neural Network solver, using a user specified ``model`` to solve a specific
    ``problem``. It can be used to solve both forward and inverse problems.

    The Residual-based Attention Physics-Informed Neural Network solver aims to
    find the solution :math:`\mathbf{u}:\Omega\rightarrow\mathbb{R}^m` of a
    differential problem:

    .. math::

        \begin{cases}
        \mathcal{A}[\mathbf{u}](\mathbf{x})=0\quad,\mathbf{x}\in\Omega\\
        \mathcal{B}[\mathbf{u}](\mathbf{x})=0\quad,
        \mathbf{x}\in\partial\Omega
        \end{cases}
    
    minimizing the loss function:

    .. math::

        \mathcal{L}_{\rm{problem}} = \frac{1}{N} \sum_{i=1}^{N_\Omega} 
        \lambda_{\Omega}^{i} \mathcal{L} \left( \mathcal{A}
        [\mathbf{u}](\mathbf{x}) \right) + \frac{1}{N} 
        \sum_{i=1}^{N_{\partial\Omega}}
        \lambda_{\partial\Omega}^{i} \mathcal{L} 
        \left( \mathcal{B}[\mathbf{u}](\mathbf{x})
        \right),
    
    denoting the weights as:
    :math:`\lambda_{\Omega}^1, \dots, \lambda_{\Omega}^{N_\Omega}` and
    :math:`\lambda_{\partial \Omega}^1, \dots, 
    \lambda_{\Omega}^{N_\partial \Omega}`
    for :math:`\Omega` and :math:`\partial \Omega`, respectively.

    Residual-based Attention Physics-Informed Neural Network updates the weights
    of the residuals at every epoch as follows:

    .. math::

        \lambda_i^{k+1} \leftarrow \gamma\lambda_i^{k} + 
        \eta\frac{\lvert r_i\rvert}{\max_j \lvert r_j\rvert},

    where :math:`r_i` denotes the residual at point :math:`i`, :math:`\gamma`
    denotes the decay rate, and :math:`\eta` is the learning rate for the
    weights' update.

    .. seealso::
        **Original reference**: Sokratis J. Anagnostopoulos, Juan D. Toscano,
        Nikolaos Stergiopulos, and George E. Karniadakis.
        *Residual-based attention and connection to information 
        bottleneck theory in PINNs.*
        Computer Methods in Applied Mechanics and Engineering 421 (2024): 116805
        DOI: `10.1016/j.cma.2024.116805
        <https://doi.org/10.1016/j.cma.2024.116805>`_.
    """

    def __init__(
        self,
        problem,
        model,
        optimizer=None,
        scheduler=None,
        weighting=None,
        loss=None,
        eta=0.001,
        gamma=0.999,
    ):
        """
        Initialization of the :class:`RBAPINN` class.

        :param AbstractProblem problem: The problem to be solved.
        :param torch.nn.Module model: The neural network model to be used.
        :param Optimizer optimizer: The optimizer to be used.
            If ``None``, the :class:`torch.optim.Adam` optimizer is used.
            Default is ``None``.
        :param Scheduler scheduler: Learning rate scheduler.
            If ``None``, the :class:`torch.optim.lr_scheduler.ConstantLR`
            scheduler is used. Default is ``None``.
        :param WeightingInterface weighting: The weighting schema to be used.
            If ``None``, no weighting schema is used. Default is ``None``.
        :param torch.nn.Module loss: The loss function to be minimized.
            If ``None``, the :class:`torch.nn.MSELoss` loss is used.
            Default is `None`.
        :param float | int eta: The learning rate for the weights of the
            residuals. Default is ``0.001``.
        :param float gamma: The decay parameter in the update of the weights
            of the residuals. Must be between ``0`` and ``1``.
            Default is ``0.999``.
        :raises: ValueError if `gamma` is not in the range (0, 1).
        :raises: ValueError if `eta` is not greater than 0.
        """
        super().__init__(
            model=model,
            problem=problem,
            optimizer=optimizer,
            scheduler=scheduler,
            weighting=weighting,
            loss=loss,
        )

        # check consistency
        check_consistency(eta, (float, int))
        check_consistency(gamma, float)

        # Validate range for gamma
        if not 0 < gamma < 1:
            raise ValueError(
                f"Invalid range: expected 0 < gamma < 1, but got {gamma}"
            )

        # Validate range for eta
        if eta <= 0:
            raise ValueError(f"Invalid range: expected eta > 0, but got {eta}")

        # Initialize parameters
        self.eta = eta
        self.gamma = gamma

        # Initialize the weight of each point to 0
        self.weights = {}
        for cond, data in self.problem.input_pts.items():
            buffer_tensor = torch.zeros((len(data), 1), device=self.device)
            self.register_buffer(f"weight_{cond}", buffer_tensor)
            self.weights[cond] = getattr(self, f"weight_{cond}")

        # Extract the reduction method from the loss function
        self._reduction = self._loss_fn.reduction

        # Set the loss function to return non-aggregated losses
        self._loss_fn = type(self._loss_fn)(reduction="none")

    def on_train_start(self):
        """
        Ensure that all residual weight buffers registered during initialization
        are moved to the correct computation device.
        """
        # Move all weight buffers to the correct device
        for cond in self.problem.input_pts:

            # Get the buffer for the current condition
            weight_buf = getattr(self, f"weight_{cond}")

            # Move the buffer to the correct device
            weight_buf.data = weight_buf.data.to(self.device)
            self.weights[cond] = weight_buf

    def training_step(self, batch, batch_idx, **kwargs):
        """
        Solver training step. It computes the optimization cycle and aggregates
        the losses using the ``weighting`` attribute.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :param int batch_idx: The index of the current batch.
        :param dict kwargs: Additional keyword arguments passed to
            ``optimization_cycle``.
        :return: The loss of the training step.
        :rtype: torch.Tensor
        """
        loss = self._optimization_cycle(
            batch=batch, batch_idx=batch_idx, **kwargs
        )
        self.store_log("train_loss", loss, self.get_batch_size(batch))
        return loss

    @torch.set_grad_enabled(True)
    def validation_step(self, batch, **kwargs):
        """
        The validation step for the PINN solver. It returns the average residual
        computed with the ``loss`` function not aggregated.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :param dict kwargs: Additional keyword arguments passed to
            ``optimization_cycle``.
        :return: The loss of the validation step.
        :rtype: torch.Tensor
        """
        losses = self.optimization_cycle(batch=batch, **kwargs)

        # Aggregate losses for each condition
        for cond, loss in losses.items():
            losses[cond] = self._apply_reduction(loss=losses[cond])

        loss = (sum(losses.values()) / len(losses)).as_subclass(torch.Tensor)
        self.store_log("val_loss", loss, self.get_batch_size(batch))
        return loss

    @torch.set_grad_enabled(True)
    def test_step(self, batch, **kwargs):
        """
        The test step for the PINN solver. It returns the average residual
        computed with the ``loss`` function not aggregated.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :param dict kwargs: Additional keyword arguments passed to
            ``optimization_cycle``.
        :return: The loss of the test step.
        :rtype: torch.Tensor
        """
        losses = self.optimization_cycle(batch=batch, **kwargs)

        # Aggregate losses for each condition
        for cond, loss in losses.items():
            losses[cond] = self._apply_reduction(loss=losses[cond])

        loss = (sum(losses.values()) / len(losses)).as_subclass(torch.Tensor)
        self.store_log("test_loss", loss, self.get_batch_size(batch))
        return loss

    def _optimization_cycle(self, batch, batch_idx, **kwargs):
        """
        Aggregate the loss for each condition in the batch.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :param int batch_idx: The index of the current batch.
        :param dict kwargs: Additional keyword arguments passed to
            ``optimization_cycle``.
        :return: The losses computed for all conditions in the batch, casted
            to a subclass of :class:`torch.Tensor`. It should return a dict
            containing the condition name and the associated scalar loss.
        :rtype: dict
        """
        # compute non-aggregated residuals
        residuals = self.optimization_cycle(batch)

        # update weights based on residuals
        self._update_weights(batch, batch_idx, residuals)

        # compute losses
        losses = {}
        for cond, res in residuals.items():

            # Get the correct indices for the weights. Modulus is used according
            # to the number of points in the condition, as in the PinaDataset.
            len_res = len(res)
            idx = torch.arange(
                batch_idx * len_res,
                (batch_idx + 1) * len_res,
                device=self.weights[cond].device,
            ) % len(self.problem.input_pts[cond])

            losses[cond] = self._apply_reduction(
                loss=(res * self.weights[cond][idx])
            )

            # store log
            self.store_log(
                f"{cond}_loss", losses[cond].item(), self.get_batch_size(batch)
            )

        # clamp unknown parameters in InverseProblem (if needed)
        self._clamp_params()

        # aggregate
        loss = self.weighting.aggregate(losses).as_subclass(torch.Tensor)

        return loss

    def _update_weights(self, batch, batch_idx, residuals):
        """
        Update weights based on residuals.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :param int batch_idx: The index of the current batch.
        :param dict residuals: A dictionary containing the residuals for each
            condition. The keys are the condition names and the values are the
            residuals as tensors.
        """
        # Iterate over each condition in the batch
        for cond, data in batch:

            # Compute normalized residuals
            res = residuals[cond]
            res_abs = torch.linalg.vector_norm(res, ord=2, dim=1, keepdim=True)
            r_norm = (self.eta * res_abs) / (res_abs.max() + 1e-12)

            # Get the correct indices for the weights. Modulus is used according
            # to the number of points in the condition, as in the PinaDataset.
            len_pts = len(data["input"])
            idx = torch.arange(
                batch_idx * len_pts,
                (batch_idx + 1) * len_pts,
                device=self.weights[cond].device,
            ) % len(self.problem.input_pts[cond])

            # Update weights
            weights = self.weights[cond]
            update = self.gamma * weights[idx] + r_norm
            weights[idx] = update.detach()

    def _apply_reduction(self, loss):
        """
        Apply the specified reduction to the loss. The reduction is deferred
        until the end of the optimization cycle to allow residual-based weights
        to be applied to each point beforehand.

        :param torch.Tensor loss: The loss tensor to be reduced.
        :return: The reduced loss tensor.
        :rtype: torch.Tensor
        :raises ValueError: If the reduction method is neither "mean" nor "sum".
        """
        # Apply the specified reduction method
        if self._reduction == "mean":
            return loss.mean()
        if self._reduction == "sum":
            return loss.sum()

        # Raise an error if the reduction method is not recognized
        raise ValueError(
            f"Unknown reduction: {self._reduction}."
            " Supported reductions are 'mean' and 'sum'."
        )



"""Module for the Interaction Network block."""

import torch
from torch_geometric.nn import MessagePassing
from ....utils import check_positive_integer
from ....model import FeedForward


class InteractionNetworkBlock(MessagePassing):
    """
    Implementation of the Interaction Network block.

    This block is used to perform message-passing between nodes and edges in a
    graph neural network, following the scheme proposed by Battaglia et al. in
    2016. It serves as an inner block in a larger graph neural network
    architecture.

    The message between two nodes connected by an edge is computed by applying a
    multi-layer perceptron (MLP) to the concatenation of the sender and
    recipient node features. Messages are then aggregated using an aggregation
    scheme (e.g., sum, mean, min, max, or product).

    The update step is performed by applying another MLP to the concatenation of
    the incoming messages and the node features.

    .. seealso::

        **Original reference**: Battaglia, P. W., et al. (2016).
        *Interaction Networks for Learning about Objects, Relations and
        Physics*.
        In Advances in Neural Information Processing Systems (NeurIPS 2016).
        DOI: `<https://doi.org/10.48550/arXiv.1612.00222>`_.
    """

    def __init__(
        self,
        node_feature_dim,
        edge_feature_dim=0,
        hidden_dim=64,
        n_message_layers=2,
        n_update_layers=2,
        activation=torch.nn.SiLU,
        aggr="add",
        node_dim=-2,
        flow="source_to_target",
    ):
        """
        Initialization of the :class:`InteractionNetworkBlock` class.

        :param int node_feature_dim: The dimension of the node features.
        :param int edge_feature_dim: The dimension of the edge features.
            If edge_attr is not provided, it is assumed to be 0.
            Default is 0.
        :param int hidden_dim: The dimension of the hidden features.
            Default is 64.
        :param int n_message_layers: The number of layers in the message
            network. Default is 2.
        :param int n_update_layers: The number of layers in the update network.
            Default is 2.
        :param torch.nn.Module activation: The activation function.
            Default is :class:`torch.nn.SiLU`.
        :param str aggr: The aggregation scheme to use for message passing.
            Available options are "add", "mean", "min", "max", "mul".
            See :class:`torch_geometric.nn.MessagePassing` for more details.
            Default is "add".
        :param int node_dim: The axis along which to propagate. Default is -2.
        :param str flow: The direction of message passing. Available options
            are "source_to_target" and "target_to_source".
            The "source_to_target" flow means that messages are sent from
            the source node to the target node, while the "target_to_source"
            flow means that messages are sent from the target node to the
            source node. See :class:`torch_geometric.nn.MessagePassing` for more
            details. Default is "source_to_target".
        :raises AssertionError: If `node_feature_dim` is not a positive integer.
        :raises AssertionError: If `hidden_dim` is not a positive integer.
        :raises AssertionError: If `n_message_layers` is not a positive integer.
        :raises AssertionError: If `n_update_layers` is not a positive integer.
        :raises AssertionError: If `edge_feature_dim` is not a non-negative
            integer.
        """
        super().__init__(aggr=aggr, node_dim=node_dim, flow=flow)

        # Check values
        check_positive_integer(node_feature_dim, strict=True)
        check_positive_integer(hidden_dim, strict=True)
        check_positive_integer(n_message_layers, strict=True)
        check_positive_integer(n_update_layers, strict=True)
        check_positive_integer(edge_feature_dim, strict=False)

        # Message network
        self.message_net = FeedForward(
            input_dimensions=2 * node_feature_dim + edge_feature_dim,
            output_dimensions=hidden_dim,
            inner_size=hidden_dim,
            n_layers=n_message_layers,
            func=activation,
        )

        # Update network
        self.update_net = FeedForward(
            input_dimensions=node_feature_dim + hidden_dim,
            output_dimensions=node_feature_dim,
            inner_size=hidden_dim,
            n_layers=n_update_layers,
            func=activation,
        )

    def forward(self, x, edge_index, edge_attr=None):
        """
        Forward pass of the block, triggering the message-passing routine.

        :param x: The node features.
        :type x: torch.Tensor | LabelTensor
        :param torch.Tensor edge_index: The edge indeces.
        :param edge_attr: The edge attributes. Default is None.
        :type edge_attr: torch.Tensor | LabelTensor
        :return: The updated node features.
        :rtype: torch.Tensor
        """
        return self.propagate(edge_index=edge_index, x=x, edge_attr=edge_attr)

    def message(self, x_i, x_j, edge_attr):
        """
        Compute the message to be passed between nodes and edges.

        :param x_i: The node features of the recipient nodes.
        :type x_i: torch.Tensor | LabelTensor
        :param x_j: The node features of the sender nodes.
        :type x_j: torch.Tensor | LabelTensor
        :return: The message to be passed.
        :rtype: torch.Tensor
        """
        if edge_attr is None:
            input_ = torch.cat((x_i, x_j), dim=-1)
        else:
            input_ = torch.cat((x_i, x_j, edge_attr), dim=-1)
        return self.message_net(input_)

    def update(self, message, x):
        """
        Update the node features with the received messages.

        :param torch.Tensor message: The message to be passed.
        :param x: The node features.
        :type x: torch.Tensor | LabelTensor
        :return: The updated node features.
        :rtype: torch.Tensor
        """
        return self.update_net(torch.cat((x, message), dim=-1))



"""Adaptive Activation Functions Module."""

__all__ = [
    "AdaptiveActivationFunctionInterface",
    "AdaptiveReLU",
    "AdaptiveSigmoid",
    "AdaptiveTanh",
    "AdaptiveSiLU",
    "AdaptiveMish",
    "AdaptiveELU",
    "AdaptiveCELU",
    "AdaptiveGELU",
    "AdaptiveSoftmin",
    "AdaptiveSoftmax",
    "AdaptiveSIREN",
    "AdaptiveExp",
]

from .adaptive_function import (
    AdaptiveReLU,
    AdaptiveSigmoid,
    AdaptiveTanh,
    AdaptiveSiLU,
    AdaptiveMish,
    AdaptiveELU,
    AdaptiveCELU,
    AdaptiveGELU,
    AdaptiveSoftmin,
    AdaptiveSoftmax,
    AdaptiveSIREN,
    AdaptiveExp,
)
from .adaptive_function_interface import AdaptiveActivationFunctionInterface



"""Module for the Equivariant Graph Neural Operator model."""

import torch
from ..utils import check_positive_integer
from .block.message_passing import EquivariantGraphNeuralOperatorBlock


class EquivariantGraphNeuralOperator(torch.nn.Module):
    """
    Equivariant Graph Neural Operator (EGNO) for modeling 3D dynamics.

    EGNO is a graph-based neural operator that preserves equivariance with
    respect to 3D transformations while modeling temporal and spatial
    interactions between nodes. It combines:

        1. Temporal convolution in the Fourier domain to capture long-range
           temporal dependencies efficiently.
        2. Equivariant Graph Neural Network (EGNN) layers to model interactions
           between nodes while respecting geometric symmetries.

    This design allows EGNO to learn complex spatiotemporal dynamics of
    physical systems, molecules, or particles while enforcing physically
    meaningful constraints.

    .. seealso::

        **Original reference**
        Xu, M., Han, J., Lou, A., Kossaifi, J., Ramanathan, A., Azizzadenesheli,
        K., Leskovec, J., Ermon, S., Anandkumar, A. (2024).
        *Equivariant Graph Neural Operator for Modeling 3D Dynamics*
        DOI: `arXiv preprint arXiv:2401.11037.
        <https://arxiv.org/abs/2401.11037>`_
    """

    def __init__(
        self,
        n_egno_layers,
        node_feature_dim,
        edge_feature_dim,
        pos_dim,
        modes,
        time_steps=2,
        hidden_dim=64,
        time_emb_dim=16,
        max_time_idx=10000,
        n_message_layers=2,
        n_update_layers=2,
        activation=torch.nn.SiLU,
        aggr="add",
        node_dim=-2,
        flow="source_to_target",
    ):
        """
        Initialization of the :class:`EquivariantGraphNeuralOperator` class.

        :param int n_egno_layers: The number of EGNO layers.
        :param int node_feature_dim: The dimension of the node features in each
            EGNO layer.
        :param int edge_feature_dim: The dimension of the edge features in each
            EGNO layer.
        :param int pos_dim: The dimension of the position features in each
            EGNO layer.
        :param int modes: The number of Fourier modes to use in the temporal
            convolution.
        :param int time_steps: The number of time steps to consider in the
            temporal convolution. Default is 2.
        :param int hidden_dim: The dimension of the hidden features in each EGNO
            layer. Default is 64.
        :param int time_emb_dim: The dimension of the sinusoidal time
            embeddings. Default is 16.
        :param int max_time_idx: The maximum time index for the sinusoidal
            embeddings. Default is 10000.
        :param int n_message_layers: The number of layers in the message
            network of each EGNO layer. Default is 2.
        :param int n_update_layers: The number of layers in the update network
            of each EGNO layer. Default is 2.
        :param torch.nn.Module activation: The activation function.
            Default is :class:`torch.nn.SiLU`.
        :param str aggr: The aggregation scheme to use for message passing.
            Available options are "add", "mean", "min", "max", "mul".
            See :class:`torch_geometric.nn.MessagePassing` for more details.
            Default is "add".
        :param int node_dim: The axis along which to propagate. Default is -2.
        :param str flow: The direction of message passing. Available options
            are "source_to_target" and "target_to_source".
            The "source_to_target" flow means that messages are sent from
            the source node to the target node, while the "target_to_source"
            flow means that messages are sent from the target node to the
            source node. See :class:`torch_geometric.nn.MessagePassing` for more
            details. Default is "source_to_target".
        :raises AssertionError: If ``n_egno_layers`` is not a positive integer.
        :raises AssertionError: If ``time_emb_dim`` is not a positive integer.
        :raises AssertionError: If ``max_time_idx`` is not a positive integer.
        :raises AssertionError: If ``time_steps`` is not a positive integer.
        """
        super().__init__()

        # Check consistency
        check_positive_integer(n_egno_layers, strict=True)
        check_positive_integer(time_emb_dim, strict=True)
        check_positive_integer(max_time_idx, strict=True)
        check_positive_integer(time_steps, strict=True)

        # Initialize parameters
        self.time_steps = time_steps
        self.time_emb_dim = time_emb_dim
        self.max_time_idx = max_time_idx

        # Initialize EGNO layers
        self.egno_layers = torch.nn.ModuleList()
        for _ in range(n_egno_layers):
            self.egno_layers.append(
                EquivariantGraphNeuralOperatorBlock(
                    node_feature_dim=node_feature_dim,
                    edge_feature_dim=edge_feature_dim,
                    pos_dim=pos_dim,
                    modes=modes,
                    hidden_dim=hidden_dim,
                    n_message_layers=n_message_layers,
                    n_update_layers=n_update_layers,
                    activation=activation,
                    aggr=aggr,
                    node_dim=node_dim,
                    flow=flow,
                )
            )

        # Linear layer to adjust the scalar feature dimension
        self.linear = torch.nn.Linear(
            node_feature_dim + time_emb_dim, node_feature_dim
        )

    def forward(self, graph):
        """
        Forward pass of the :class:`EquivariantGraphNeuralOperator` class.

        :param graph: The input graph object with the following attributes:
            - 'x': Node features, shape ``[num_nodes, node_feature_dim]``.
            - 'pos': Node positions, shape ``[num_nodes, pos_dim]``.
            - 'vel': Node velocities, shape ``[num_nodes, pos_dim]``.
            - 'edge_index': Graph connectivity, shape ``[2, num_edges]``.
            - 'edge_attr': Edge attrs, shape ``[num_edges, edge_feature_dim]``.
        :type graph: Data | Graph
        :return: The output graph object with updated node features,
            positions, and velocities. The output graph adds to 'x', 'pos',
            'vel', and 'edge_attr' the time dimension, resulting in shapes:
            - 'x': ``[time_steps, num_nodes, node_feature_dim]``
            - 'pos': ``[time_steps, num_nodes, pos_dim]``
            - 'vel': ``[time_steps, num_nodes, pos_dim]``
            - 'edge_attr': ``[time_steps, num_edges, edge_feature_dim]``
        :rtype: Data | Graph
        :raises ValueError: If the input graph does not have a 'vel' attribute.
        """
        # Check that the graph has the required attributes
        if "vel" not in graph:
            raise ValueError("The input graph must have a 'vel' attribute.")

        # Compute the temporal embedding
        emb = self._embedding(torch.arange(self.time_steps)).to(graph.x.device)
        emb = emb.unsqueeze(1).repeat(1, graph.x.shape[0], 1)

        # Expand dimensions
        x = graph.x.unsqueeze(0).repeat(self.time_steps, 1, 1)
        x = self.linear(torch.cat((x, emb), dim=-1))
        pos = graph.pos.unsqueeze(0).repeat(self.time_steps, 1, 1)
        vel = graph.vel.unsqueeze(0).repeat(self.time_steps, 1, 1)

        # Manage edge index
        offset = torch.arange(self.time_steps).reshape(-1, 1)
        offset = offset.to(graph.x.device) * graph.x.shape[0]
        src = graph.edge_index[0].unsqueeze(0) + offset
        dst = graph.edge_index[1].unsqueeze(0) + offset
        edge_index = torch.stack([src, dst], dim=0).reshape(2, -1)

        # Manage edge attributes
        if graph.edge_attr is not None:
            edge_attr = graph.edge_attr.unsqueeze(0)
            edge_attr = edge_attr.repeat(self.time_steps, 1, 1)
        else:
            edge_attr = None

        # Iteratively apply EGNO layers
        for layer in self.egno_layers:
            x, pos, vel = layer(
                x=x,
                pos=pos,
                vel=vel,
                edge_index=edge_index,
                edge_attr=edge_attr,
            )

        # Build new graph
        new_graph = graph.clone()
        new_graph.x, new_graph.pos, new_graph.vel = x, pos, vel
        if edge_attr is not None:
            new_graph.edge_attr = edge_attr

        return new_graph

    def _embedding(self, time):
        """
        Generate sinusoidal temporal embeddings.

        :param torch.Tensor time: The time instances.
        :return: The sinusoidal embedding tensor.
        :rtype: torch.Tensor
        """
        # Compute the sinusoidal embeddings
        half_dim = self.time_emb_dim // 2
        logs = torch.log(torch.as_tensor(self.max_time_idx)) / (half_dim - 1)
        freqs = torch.exp(-torch.arange(half_dim) * logs)
        args = torch.as_tensor(time)[:, None] * freqs[None, :]
        emb = torch.cat([torch.sin(args), torch.cos(args)], dim=-1)

        # Apply padding if the embedding dimension is odd
        if self.time_emb_dim % 2 == 1:
            emb = torch.nn.functional.pad(emb, (0, 1), mode="constant")

        return emb



"""Module for the DataCondition class."""

import torch
from torch_geometric.data import Data
from .condition_interface import ConditionInterface
from ..label_tensor import LabelTensor
from ..graph import Graph


class DataCondition(ConditionInterface):
    """
    The class :class:`DataCondition` defines an unsupervised condition based on
    ``input`` data. This condition is typically used in data-driven problems,
    where the model is trained using a custom unsupervised loss determined by
    the chosen :class:`~pina.solver.solver.SolverInterface`, while leveraging
    the provided data during training. Optional ``conditional_variables`` can be
    specified when the model depends on additional parameters.

    The class automatically selects the appropriate implementation based on the
    type of the ``input`` data. Depending on whether the ``input`` is a tensor
    or graph-based data, one of the following specialized subclasses is
    instantiated:

    - :class:`TensorDataCondition`: For cases where the ``input`` is either a
      :class:`torch.Tensor` or a :class:`~pina.label_tensor.LabelTensor` object.

    - :class:`GraphDataCondition`: For cases where the ``input`` is either a
      :class:`~pina.graph.Graph` or :class:`~torch_geometric.data.Data` object.

    :Example:

    >>> from pina import Condition, LabelTensor
    >>> import torch

    >>> pts = LabelTensor(torch.randn(100, 2), labels=["x", "y"])
    >>> cond_vars = LabelTensor(torch.randn(100, 1), labels=["w"])
    >>> condition = Condition(input=pts, conditional_variables=cond_vars)
    """

    # Available input data types
    __slots__ = ["input", "conditional_variables"]
    _avail_input_cls = (torch.Tensor, LabelTensor, Data, Graph, list, tuple)
    _avail_conditional_variables_cls = (torch.Tensor, LabelTensor)

    def __new__(cls, input, conditional_variables=None):
        """
        Instantiate the appropriate subclass of :class:`DataCondition` based on
        the type of the ``input``.

        :param input: The input data for the condition.
        :type input: torch.Tensor | LabelTensor | Graph |
            Data | list[Graph] | list[Data] | tuple[Graph] | tuple[Data]
        :param conditional_variables: The conditional variables for the
            condition. Default is ``None``.
        :type conditional_variables: torch.Tensor | LabelTensor
        :return: The subclass of DataCondition.
        :rtype: pina.condition.data_condition.TensorDataCondition |
            pina.condition.data_condition.GraphDataCondition
        :raises ValueError: If ``input`` is not of type :class:`torch.Tensor`,
            :class:`~pina.label_tensor.LabelTensor`, :class:`~pina.graph.Graph`,
            or :class:`~torch_geometric.data.Data`.
        """
        if cls != DataCondition:
            return super().__new__(cls)

        # If the input is a tensor
        if isinstance(input, (torch.Tensor, LabelTensor)):
            subclass = TensorDataCondition
            return subclass.__new__(subclass, input, conditional_variables)

        # If the input is a graph
        if isinstance(input, (Graph, Data, list, tuple)):
            cls._check_graph_list_consistency(input)
            subclass = GraphDataCondition
            return subclass.__new__(subclass, input, conditional_variables)

        # If the input is not of the correct type raise an error
        raise ValueError(
            "Invalid input types. "
            "Please provide either torch_geometric.data.Data or Graph objects."
        )

    def __init__(self, input, conditional_variables=None):
        """
        Initialization of the :class:`DataCondition` class.

        :param input: The input data for the condition.
        :type input: torch.Tensor | LabelTensor | Graph | Data | list[Graph] |
            list[Data] | tuple[Graph] | tuple[Data]
        :param conditional_variables: The conditional variables for the
            condition. Default is ``None``.
        :type conditional_variables: torch.Tensor | LabelTensor

        .. note::

            If ``input`` is a list of :class:`~pina.graph.Graph` or
            :class:`~torch_geometric.data.Data`, all elements in
            the list must share the same structure, with matching keys and
            consistent data types.
        """
        super().__init__()
        self.input = input
        self.conditional_variables = conditional_variables


class TensorDataCondition(DataCondition):
    """
    Specialization of the :class:`DataCondition` class for the case where
    ``input`` is either a :class:`~pina.label_tensor.LabelTensor` object or a
    :class:`torch.Tensor` object.
    """


class GraphDataCondition(DataCondition):
    """
    Specialization of the :class:`DataCondition` class for the case where
    ``input`` is either a :class:`~pina.graph.Graph` object or a
    :class:`~torch_geometric.data.Data` object.
    """



"""Module for the Exclusion Operation."""

import random
import torch
from ..label_tensor import LabelTensor
from .operation_interface import OperationInterface


class Exclusion(OperationInterface):
    r"""
    Implementation of the exclusion operation between of a list of domains.

    Given two sets :math:`A` and :math:`B`, define the exclusion of the two
    sets as:

    .. math::
        A \setminus B = \{x \mid x \in A \land x \in B \land
        x \not\in(A \lor B)\},

    where :math:`x` is a point in :math:`\mathbb{R}^N`.
    """

    def __init__(self, geometries):
        """
        Initialization of the :class:`Exclusion` class.

        :param list[DomainInterface] geometries: A list of instances of the
            :class:`~pina.domain.domain_interface.DomainInterface` class on
            which the exclusion operation is performed.

        :Example:
            >>> # Create two ellipsoid domains
            >>> ellipsoid1 = EllipsoidDomain({'x': [-1, 1], 'y': [-1, 1]})
            >>> ellipsoid2 = EllipsoidDomain({'x': [0, 2], 'y': [0, 2]})
            >>> # Define the exclusion between the domains
            >>> exclusion = Exclusion([ellipsoid1, ellipsoid2])
        """
        super().__init__(geometries)

    def is_inside(self, point, check_border=False):
        """
        Check if a point is inside the resulting domain.

        :param LabelTensor point: Point to be checked.
        :param bool check_border: If ``True``, the border is considered inside
            the domain. Default is ``False``.
        :return: ``True`` if the point is inside the domain,
            ``False`` otherwise.
        :rtype: bool
        """
        flag = 0
        for geometry in self.geometries:
            if geometry.is_inside(point, check_border):
                flag += 1
        return flag == 1

    def sample(self, n, mode="random", variables="all"):
        """
        Sampling routine.

        :param int n: Number of points to sample.
        :param str mode: Sampling method. Default is ``random``.
            Available modes: random sampling, ``random``;
        :param list[str] variables: variables to be sampled. Default is ``all``.
        :raises NotImplementedError: If the sampling method is not implemented.
        :return: Sampled points.
        :rtype: LabelTensor

        :Example:
            >>> # Create two Cartesian domains
            >>> cartesian1 = CartesianDomain({'x': [0, 2], 'y': [0, 2]})
            >>> cartesian2 = CartesianDomain({'x': [1, 3], 'y': [1, 3]})
            >>> # Define the exclusion between the domains
            >>> Exclusion = Exclusion([cartesian1, cartesian2])
            >>> # Sample
            >>> Exclusion.sample(n=5)
                LabelTensor([[2.4187, 1.5792],
                            [2.7456, 2.3868],
                            [2.3830, 1.7037],
                            [0.8636, 1.8453],
                            [0.1978, 0.3526]])
            >>> len(Exclusion.sample(n=5)
                5
        """
        if mode not in self.sample_modes:
            raise NotImplementedError(
                f"{mode} is not a valid mode for sampling."
            )

        sampled = []

        # calculate the number of points to sample for each geometry and the
        # remainder.
        remainder = n % len(self.geometries)
        num_points = n // len(self.geometries)

        # sample the points
        # NB. geometries as shuffled since if we sample
        # multiple times just one point, we would end
        # up sampling only from the first geometry.
        iter_ = random.sample(self.geometries, len(self.geometries))
        for i, geometry in enumerate(iter_):
            sampled_points = []
            # int(i < remainder) is one only if we have a remainder
            # different than zero. Notice that len(geometries) is
            # always smaller than remaider.
            # makes sure point is uniquely inside 1 shape.
            while len(sampled_points) < (num_points + int(i < remainder)):
                sample = geometry.sample(1, mode, variables)
                # if not self.is_inside(sample) --> will be the intersection
                if self.is_inside(sample):
                    sampled_points.append(sample)
            sampled += sampled_points

        return LabelTensor(torch.cat(sampled), labels=self.variables)
